import numpy as np
import scipy

from .algorithm_utils import (
    ParameterSerializer,
    Problem,
    get_bias_configuration,
    get_metric_mode,
    init_random_parameters,
    normalize_propensities,
    predict_scores,
    process_propensities,
)


class DebiasingLearningAlgorithm:
    def __init__(
        self, K, normalization, lamda, metric, bias_mode, clip_val=-1, verbose=0
    ):
        self.normalization = normalization
        self.metric = metric
        self.K = K
        self.clip_val = clip_val
        self.lamda = lamda
        self.bias_mode = bias_mode
        self.verbose = verbose

    def train(self, train_observations, inv_propensities, start_vector=None):
        if start_vector is not None:
            assert start_vector[0].shape[1] == self.K
            assert start_vector[1].shape[1] == self.K

        tempInvPropensities = None
        if inv_propensities is not None:
            tempInvPropensities = (4.0 / 3.0) * inv_propensities
            if self.clip_val >= 0:
                tempInvPropensities = np.clip(
                    tempInvPropensities, a_min=0, a_max=self.clip_val
                )

        parameters = self.optimize_debiasing_parameters(
            train_observations, tempInvPropensities, self.lamda, start_vec=start_vector
        )
        return parameters

    def optimize_debiasing_parameters(
        self, observed_ratings, inverse_propensities, l2_regularization, start_vec=None
    ):

        metricMode = get_metric_mode(self.metric)
        numUsers, numItems = np.shape(observed_ratings)
        scale = numUsers * numItems
        useBias, regularizeBias = get_bias_configuration(self.bias_mode)
        serializer = ParameterSerializer(
            numUsers=numUsers, numItems=numItems, num_dimensions=self.K
        )
        config = {
            "numUsers": numUsers,
            "numItems": numItems,
            "scale": scale,
            "useBias": useBias,
            "regularizeBias": regularizeBias,
            "metricMode": metricMode,
            "num_dimensions": self.K,
            "l2_regularization": l2_regularization,
            "serializer": serializer,
        }

        inversePropensities = process_propensities(
            observed_ratings, inverse_propensities
        )
        normalized_propensities = normalize_propensities(
            inversePropensities, self.normalization, scale, numItems, numUsers
        )
        normalized_propensities = np.ma.filled(normalized_propensities, 0.0)
        observed_ratings = np.ma.filled(observed_ratings, 0)

        problem = Problem(
            observed_ratings=observed_ratings,
            normalized_propensities=normalized_propensities,
            config=config,
        )

        startparameters = init_random_parameters(config, start_vec=start_vec)
        startVector = serializer.serialize(*startparameters)

        ops = {
            "maxiter": 2000,
            "disp": False,
            "gtol": 1e-5,
            "ftol": 1e-5,
            "maxcor": 50,
            "disp": False,
        }

        result = scipy.optimize.minimize(
            fun=problem.evaluate,
            x0=startVector,
            method="L-BFGS-B",
            jac=True,
            tol=1e-5,
            options=ops,
        )

        return serializer.deserialize(result["x"])

    def predict(self, model_params, bias_mode):
        use_bias = True
        if bias_mode == "None":
            use_bias = False
        return predict_scores(*model_params, use_bias=use_bias)



import numpy as np


def distance(a, b, order=2):
    """
    Description
    -----------
    Calculates the specified norm between two vectors.

    Parameters
    ----------
        a : list
            First vector
        b : list
            Second vector:
        order :  int
            Order of the norm to be calculated as distance

    Returns
    -------
        Resultant norm value
    """
    assert len(a) == len(b), "Length of the vectors for distance don't match."
    return np.linalg.norm(x=np.array(a) - np.array(b), ord=order)



import numpy as np
from tqdm import trange

from .algorithm_utils import (
    exposure_diff,
    find_items_per_group_per_query,
    hh,
    normalized_topp_prot_deriv_per_group,
    topp,
)


class DELTRAlgorithm(object):
    def __init__(
        self, gamma, number_of_iterations, learning_rate, lambdaa, init_var, verbose
    ):
        # assign parameters
        self._gamma = gamma
        self._number_of_iterations = number_of_iterations
        self._learning_rate = learning_rate
        self._lambda = lambdaa
        self._init_var = init_var
        self.verbose = verbose

        self._no_exposure = False
        if gamma == 0:
            self._no_exposure = True

        self.log = []

        self._data_per_query = {}

    def fit(self, query_ids, protected_feature, training_scores, feature_matrix):
        """
        Description
        -----------
        Train a Neural Network to find the optimal feature weights in listwise learning to rank.

        Parameters
        ----------
        query_ids: list
            List of query IDs.

        feature_matrix: matrix-like
            training features.

        training_scores: array-like
            training judgments.

        Return
        ------
            model parameters.
        """
        self.indexes_per_query = {
            q: np.where(np.array(query_ids == q))[0] for q in query_ids.unique()
        }
        self.queries = list(self.indexes_per_query.keys())

        n_features = feature_matrix.shape[1]
        for q, indexes in self.indexes_per_query.items():
            self._data_per_query[
                hh(q, training_scores)
            ] = find_items_per_group_per_query(
                training_scores.iloc[indexes], protected_feature.iloc[indexes]
            )

            self._data_per_query[
                hh(q, feature_matrix)
            ] = find_items_per_group_per_query(
                feature_matrix.iloc[indexes], protected_feature.iloc[indexes]
            )
        self.feature_data_per_query_list = []
        self.true_data_per_query_list = []
        self.prot_idx_per_query_list = []
        for query in self.queries:
            feature_data_per_query = self._data_per_query[hh(query, feature_matrix)][0]
            true_data_per_query = self._data_per_query[hh(query, training_scores)][0]
            prot_idx_per_query = protected_feature.iloc[self.indexes_per_query[query]]

            self.feature_data_per_query_list.append(feature_data_per_query)
            self.true_data_per_query_list.append(true_data_per_query)
            self.prot_idx_per_query_list.append(prot_idx_per_query)

        # linear neural network parameter initialization
        omega = (np.random.rand(n_features, 1) * self._init_var).reshape(-1)
        for t in trange(self._number_of_iterations, leave=self.verbose > 0):
            # forward propagation
            predicted_scores = np.dot(feature_matrix, omega)
            predicted_scores = np.reshape(
                predicted_scores, (feature_matrix.shape[0], 1)
            )
            grad = self._calculate_gradient(predicted_scores)
            omega = omega - self._learning_rate * np.sum(
                np.asarray(grad), axis=0
            ).reshape(-1)

        return omega

    def _calculate_cost(
        self, training_judgments, predictions, prot_idx, data_per_query_predicted
    ):
        """
        Description
        -----------
        Computes the loss in list-wise learning to rank.

        Parameters
        ----------
        training_judgments: pd.Series
            containing the training judgments/ scores.

        predictions: pd.Series
            containing the predicted scores.

        prot_idx: pd.Series
            list stating which item is protected or non-protected.

        data_per_query_predicted:
            stores all judgments and all predicted scores that belong to one query.

        Return
        ------
            float value --> loss
        """
        results = [
            self._loss(
                query,
                training_judgments,
                predictions,
                prot_idx,
                data_per_query_predicted,
            )
            for query in self.queries
        ]

        # calucalte losses for better debugging
        loss_standard = sum(results)[0]
        loss_exposure = sum(
            [
                exposure_diff(predictions[indexes], prot_idx[indexes])
                for indexes in self.indexes_per_query.values()
            ]
        )

        return np.asarray(results), loss_standard, loss_exposure

    def _loss(
        self,
        which_query,
        training_judgments,
        predictions,
        prot_idx,
        data_per_query_predicted,
    ):
        """Calculate loss for a given query"""

        result = -np.dot(
            np.transpose(
                topp(self._data_per_query[(hh(which_query, training_judgments))][0])
            ),
            np.log(topp(data_per_query_predicted[hh(which_query, predictions)][0])),
        ) / np.log(predictions.size)

        if not self._no_exposure:
            indexes = self.indexes_per_query[which_query]
            result += (
                self._gamma
                * exposure_diff(predictions[indexes], prot_idx[indexes]) ** 2
            )

        return result

    def _calculate_gradient(self, predictions):
        """
        Description
        -----------
        calculates local gradients of current feature weights.

        Parameters
        ----------
        predictions:
            Vector containing the prediction scores.

        Return
        ------
            float value --> optimal listwise cost.
        """
        # Pool
        # from multiprocessing import Pool
        # with Pool(4) as p:
        #    results = list(p.starmap(self._grad, [(query, training_features, training_judgments, predictions, prot_idx,
        #                      data_per_query_predicted) for query in self.queries]))

        # Joblib
        # results = Parallel(n_jobs=4, verbose=0)(
        #   delayed(self._grad)(query, training_features, training_judgments, predictions, prot_idx,
        #                      data_per_query_predicted) for query in self.queries
        # )

        predictions_per_query_list = []
        for query, prot_idx_per_query in zip(
            self.queries, self.prot_idx_per_query_list
        ):
            predictions_per_query = predictions[self.indexes_per_query[query]].astype(
                "float32"
            )
            predictions_per_query = np.exp(
                find_items_per_group_per_query(
                    predictions_per_query, prot_idx_per_query
                )[0]
            )
            predictions_per_query_list.append(predictions_per_query)

        args_list = (
            self.feature_data_per_query_list,
            predictions_per_query_list,
            self.prot_idx_per_query_list,
            self.true_data_per_query_list,
        )
        results = [self._grad(predictions.size, *args) for args in zip(*args_list)]
        return np.asarray(results)

    def _grad(
        self,
        predictions_size,
        feature_data_per_query,
        predictions_per_query,
        prot_idx_per_query,
        true_data_per_query,
    ):

        feature_data_per_query_transpose = np.transpose(feature_data_per_query)

        result = 1 / np.sum(predictions_per_query)
        # l3
        result *= np.dot(feature_data_per_query_transpose, predictions_per_query)
        # l1
        result += -np.dot(feature_data_per_query_transpose, topp(true_data_per_query))
        # L deriv
        result /= np.log(predictions_size)

        if not self._no_exposure:
            result = result.reshape(-1) + self._gamma * 2 * exposure_diff(
                predictions_per_query, prot_idx_per_query
            ) * self._normalized_topp_prot_deriv_per_group_diff(
                feature_data_per_query, predictions_per_query, prot_idx_per_query
            )
        return result

    def _normalized_topp_prot_deriv_per_group_diff(
        self, training_features_per_query, predictions_per_query, prot_idx_per_query
    ):
        """
        Description
        -----------
        Calculates the difference of the normalized topp_prot derivative of the protected and non-protected groups.

        Parameters
        ----------
        training_features:
            vector of all features.

        predictions:
            predictions of all data points

        prot_idx:
            list stating which item is protected or non-protected.

        Return
        ------
            numpy array of float values.
        """

        (
            train_judgments_per_query,
            train_protected_items_per_query,
            train_nonprotected_items_per_query,
        ) = find_items_per_group_per_query(
            training_features_per_query, prot_idx_per_query
        )

        (
            predictions_per_query,
            pred_protected_items_per_query,
            pred_nonprotected_items_per_query,
        ) = find_items_per_group_per_query(predictions_per_query, prot_idx_per_query)

        u2 = normalized_topp_prot_deriv_per_group(
            train_nonprotected_items_per_query,
            train_judgments_per_query,
            pred_nonprotected_items_per_query,
            predictions_per_query,
        )  # derivative for non-protected group
        u3 = normalized_topp_prot_deriv_per_group(
            train_protected_items_per_query,
            train_judgments_per_query,
            pred_protected_items_per_query,
            predictions_per_query,
        )  # derivative for protected group

        return u2 - u3

    def losses(self):
        return self.log



# Base Imports
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt

from holisticai.utils import get_colors

# utils
from ...utils._validation import _check_binary, _multiclass_checks

# Import metrics
from ..metrics import frequency_matrix


def frequency_plot(p_attr, y_pred, ax=None, size=None, title=None):
    """
    Frequency plot.

    Description
    ----------
    This function plots how frequently members
    of each group fall into each class.

    Parameters
    ----------
    p_attr : array-like
        Protected attribute vector
    y_pred : array-like

    Returns
    -------
    matplotlib ax (or None)
    """
    # check and coerce inputs
    p_attr, y_pred, _, _, _ = _multiclass_checks(
        p_attr=p_attr,
        y_pred=y_pred,
        y_true=None,
        groups=None,
        classes=None,
    )

    # get success rates
    sr_list = frequency_matrix(p_attr, y_pred, normalize="group")

    # sort by success rate
    sr_tot = sr_list.sum(axis=0) / sr_list.sum(axis=0).sum()
    sr_tot.name = "total"
    sr_list = pd.concat([sr_list, pd.DataFrame(sr_tot).transpose()], axis=0)
    name_classes = sr_list.columns
    n_classes = len(name_classes)

    # charting
    sns.set()

    colors = get_colors(sr_list.shape[0])
    hai_palette = sns.color_palette(colors)

    if n_classes > 2:
        for i in range(n_classes):
            fig, ax = plt.subplots()
            fig.suptitle("Class " + str(name_classes[i]))
            sns.barplot(
                x=sr_list.index.to_list(),
                y=sr_list[name_classes[i]],
                palette=hai_palette,
                ax=ax,
            )
            ax.set_xlabel("Group")
            ax.set_ylabel("Frequency ")
            _, labels = plt.xticks()
            plt.setp(labels, rotation=45)

        return None

    else:

        if ax is None:
            fig, ax = plt.subplots(figsize=size)
            if title is not None:
                fig.suptitle(title)
            else:
                fig.suptitle("Frequency Plot (Class {})".format(name_classes[1]))

        sns.barplot(
            x=sr_list.index.to_list(),
            y=sr_list[name_classes[1]],
            palette=hai_palette,
            ax=ax,
        )
        ax.set_xlabel("Group")
        ax.set_ylabel("Frequency")
        _, labels = plt.xticks()
        plt.setp(labels, rotation=45)

        return ax


def statistical_parity_plot(
    p_attr, y_pred, pos_label=1, compare_to=None, ax=None, size=None, title=None
):
    """
    Statistical Parity Plot (Binary Classification).

    Description
    -----------
    This function plots the statistical parity for each group
    along with acceptable bounds. We take the group with maximum
    success rate as the comparison group.

    Parameters
    ----------
    p_attr : array-like
        Protected attribute vector
    y_pred : array-like
        Prediction vector (binary)
    pos_label (optional) : label, default=1
        The positive label
    compare_to (optional) : str or int
        The group we are comparing to
    ax (optional) : matplotlib axes
        Pre-existing axes for the plot
    size (optional) : (int, int)
        Size of the figure
    title (optional) : str
        Title of the figure

    Returns
    -------
    matplotlib ax
    """
    # check and coerce inputs
    p_attr, y_pred, _, groups, _ = _multiclass_checks(
        p_attr=p_attr,
        y_pred=y_pred,
        y_true=None,
        groups=None,
        classes=None,
    )

    _check_binary(y_pred, name="y_pred")

    group_dict = dict(zip(groups, range(len(groups))))
    # get success rates.
    sr_list = frequency_matrix(p_attr, y_pred * 1, groups=groups, normalize="group")[
        pos_label
    ].to_numpy()

    # sort by success rate.
    sr_list_sorted, groups_sorted = zip(*sorted(zip(sr_list, groups), reverse=True))
    sr_list_sorted, groups_sorted = list(sr_list_sorted), list(groups_sorted)

    # statistical parity list
    if compare_to is not None:
        sp_list = sr_list_sorted - sr_list[group_dict[compare_to]]
    else:
        sp_list = sr_list_sorted - np.max(sr_list)

    # setup
    sns.set()
    if ax is None:
        fig, ax = plt.subplots(figsize=size)
        if title is not None:
            fig.suptitle(title)
        else:
            fig.suptitle("Statistical Parity plot")

    # charting
    colors = get_colors(len(groups))
    hai_palette = sns.color_palette(colors)
    ax.set_xlabel("Group")
    ax.set_ylabel("Statistical Parity")
    sns.barplot(x=groups_sorted, y=sp_list, palette=hai_palette, ax=ax)
    # horizontal lines
    ax.axhline(y=-0.1, color="grey", linestyle="--", label="lower bound")
    ax.axhline(y=0.1, color="grey", linestyle="--", label="upper bound")
    ax.axhspan(-0.1, 0.1, alpha=0.3, color="grey", zorder=0, label="fair area")
    # tilt labels
    _, labels = plt.xticks()
    plt.setp(labels, rotation=45)
    ax.legend()

    return ax


def disparate_impact_plot(
    p_attr, y_pred, pos_label=1, compare_to=None, ax=None, size=None, title=None
):
    """
    Disparate Impact Plot (Binary Classification).

    Description
    -----------
    This function plots the disparate impact for each group
    along with acceptable bounds. We take the group with maximum
    success rate as the 'majority group'.

    Parameters
    ----------
    p_attr : array-like
        Protected attribute vector
    y_pred : array-like
        Prediction vector
    pos_label : label, default=1
        The positive label
    compare_to (optional) : str or int
        The group we are comparing to
    ax (optional) : matplotlib axes
        Pre-existing axes for the plot
    size (optional) : (int, int)
        Size of the figure
    title (optional) : str
        Title of the figure

    Returns
    -------
    matplotlib ax
    """
    # check and coerce inputs
    p_attr, y_pred, _, groups, _ = _multiclass_checks(
        p_attr=p_attr,
        y_pred=y_pred,
        y_true=None,
        groups=None,
        classes=None,
    )

    _check_binary(y_pred, name="y_pred")

    group_dict = dict(zip(groups, range(len(groups))))
    # get success rates
    sr_list = frequency_matrix(p_attr, 1 * y_pred, groups)[pos_label].to_numpy()

    # sort by success rate.
    sr_list_sorted, groups_sorted = zip(*sorted(zip(sr_list, groups), reverse=True))
    sr_list_sorted, groups_sorted = list(sr_list_sorted), list(groups_sorted)

    # disparate impact list
    if compare_to is not None:
        di_list = sr_list_sorted / sr_list[group_dict[compare_to]]
    else:
        di_list = sr_list_sorted / sr_list_sorted[0]

    # setup
    sns.set()
    if ax is None:
        fig, ax = plt.subplots(figsize=size)
        if title is not None:
            fig.suptitle(title)
        else:
            fig.suptitle("Disparate Impact plot")

    # charting
    colors = get_colors(len(groups))
    hai_palette = sns.color_palette(colors)
    ax.set_xlabel("Group")
    ax.set_ylabel("Disparate Impact")
    sns.barplot(x=groups_sorted, y=di_list, palette=hai_palette, ax=ax)
    # horizontal lines
    ax.axhspan(0.8, 1.2, alpha=0.3, color="grey", label="fair area")
    ax.axhline(y=1.2, color="grey", linestyle="--", label="upper bound")
    ax.axhline(y=0.8, color="grey", linestyle="--", label="lower bound")
    # tilt labels
    _, labels = plt.xticks()
    plt.setp(labels, rotation=45)
    # legend
    ax.legend()

    # return
    return ax


def frequency_matrix_plot(
    p_attr,
    y_pred,
    groups=None,
    classes=None,
    normalize=None,
    reverse_colors=False,
    ax=None,
    size=None,
    title=None,
):
    """
    Frequency Matrix Plot.

    Description
    -----------
    This function plots the matrix of occurence rate (count)
    for each group, class pair. We include the option to normalise
    over groups or classes.

    Parameters
    ----------
    p_attr : array-like
        Protected attribute vector
    y_pred : array-like
        Prediction vector (categorical)
    groups (optional) : array or list
        The groups in order
    classes (optional) : array or list
        The classes in order
    normalize (optional): None, 'group' or 'class'
        According to which of group or class we normalize
    reverse_colors (optional): bool, default=False
        Option to reverse the color palette
    ax (optional) : matplotlib axes
        Pre-existing axes for the plot
    size (optional) : (int, int)
        Size of the figure
    title (optional) : str
        Title of the figure

    Returns
    -------
    matplotlib ax
    """
    # check and coerce inputs
    p_attr, y_pred, _, groups, classes = _multiclass_checks(
        p_attr=p_attr,
        y_pred=y_pred,
        y_true=None,
        groups=groups,
        classes=classes,
    )

    # compute frequency matrix
    colors = get_colors(10, extended_colors=True, reverse=reverse_colors)
    hai_palette = sns.color_palette(colors)
    sr_mat = frequency_matrix(
        p_attr, y_pred, groups=groups, classes=classes, normalize=normalize
    )

    # setup
    sns.set()
    if ax is None:
        fig, ax = plt.subplots(figsize=size)
        if title is not None:
            fig.suptitle(title)
        else:
            fig.suptitle("Frequency matrix plot")

    # charting
    if normalize is None:
        sns.heatmap(sr_mat, annot=True, cmap=hai_palette, ax=ax)
    else:
        sns.heatmap(sr_mat, annot=True, fmt=".2%", cmap=hai_palette, ax=ax)
    ax.set_xlabel("Class")
    ax.set_ylabel("Group")

    # return
    return ax


def accuracy_bar_plot(p_attr, y_pred, y_true, ax=None, size=None, title=None):
    """
    Accuracy Bar Plot.

    Description
    -----------
    This function plots the accuracy of the predictions
    for each group.

    Parameters
    ----------
    p_attr : array-like
        Protected attribute vector
    y_pred : array-like
        Prediction vector
    y_true : array-like
        Target vector
    ax (optional) : matplotlib axes
        Pre-existing axes for the plot
    size (optional) : (int, int)
        Size of the figure
    title (optional) : str
        Title of the figure

    Returns
    -------
    matplotlib ax
    """
    # check and coerce inputs
    p_attr, y_pred, y_true, groups, _ = _multiclass_checks(
        p_attr=p_attr,
        y_pred=y_pred,
        y_true=y_true,
        groups=None,
        classes=None,
    )

    # loop over groups
    acc_list = []
    for c in groups:
        members = p_attr == c
        pred_c = y_pred[members]
        true_c = y_true[members]
        truepred = (pred_c == true_c).sum()
        acc = truepred / len(pred_c)
        acc_list.append(acc)

    acc_list = list(acc_list)
    groups = list(groups)

    # tot
    acc_tot = (y_pred == y_true).sum() / len(y_pred)
    acc_list.append(acc_tot)
    groups.append("Total")

    # setup
    sns.set()
    if ax is None:
        fig, ax = plt.subplots(figsize=size)
        if title is not None:
            fig.suptitle(title)
        else:
            fig.suptitle("Accuracy Bar Plot")

    # charting
    colors = get_colors(len(groups))
    hai_palette = sns.color_palette(colors)
    ax.set_xlabel("Group")
    ax.set_ylabel("Accuracy")
    sns.barplot(x=groups, y=acc_list, palette=hai_palette, ax=ax)

    return ax



import pandas as pd


class SensitiveGroups:
    def merge_columns(self, feature_columns):
        return pd.DataFrame(feature_columns).apply(
            lambda row: ",".join([str(r) for r in row.values]), axis=1
        )

    def fit(self, sensitive_features):
        self.tags = pd.DataFrame()
        group_lbs = self.merge_columns(sensitive_features)
        _, unique_group_lbs = pd.factorize(group_lbs)
        self.num_groups = len(unique_group_lbs)
        self.group_names = list(unique_group_lbs)
        self.group2num = {g: i for i, g in enumerate(unique_group_lbs)}
        return self

    def transform(self, sensitive_features, convert_numeric=False):
        group_lbs = self.merge_columns(sensitive_features)
        if convert_numeric:
            return group_lbs.apply(lambda x: self.group2num[x])
        return group_lbs

    def fit_transform(self, sensitive_features, convert_numeric=False):
        return self.fit(sensitive_features).transform(
            sensitive_features, convert_numeric=convert_numeric
        )



import sys
import time
from collections import OrderedDict
from time import gmtime, strftime


class Logging:
    def __init__(
        self, log_params, total_iterations=None, epochs=None, logger_format="iteration"
    ):
        self.epochs = epochs
        self.total_iterations = total_iterations
        self.logger_format = logger_format
        self.log_params = log_params
        self.time_start = time.time()

    def update(self, *args, **kargs):
        init = False
        finish = False
        params = OrderedDict()
        for param_value, (param_name, param_type) in zip(args, self.log_params):
            if param_name == "iteration":
                if self.logger_format == "iteration":
                    params["iter"] = f"{int(param_value)}/{int(self.total_iterations)}"
                    init = int(param_value) == 0
                    finish = int(param_value) == int(self.total_iterations)
                elif self.logger_format == "epochs":
                    epoch = (param_value / self.total_iterations) * self.epochs
                    params["epochs"] = f"{ epoch: .2f}"
                    init = param_value == 0
                    finish = int(param_value) == int(self.total_iterations)

            elif param_type is int:
                params[param_name] = param_value

            elif param_type is float:
                params[param_name] = f"{param_value:.4f}"

            elif param_type is str:
                params[param_name] = f"{param_value}"

        current_time = time.time()
        elapsed_time = current_time - self.time_start
        elapsed_ftime = strftime("%H:%M:%S", gmtime(elapsed_time))

        content = f"elapsed time: {elapsed_ftime}"
        for key, value in params.items():
            content += f" | {key}:{value}"
        content = f"[{content}]\r"

        if init:
            content = f"\n{content}"

        if finish:
            content = f"{content}\n"

        sys.stdout.write(content)
        sys.stdout.flush()

    def info(self, message):
        current_time = time.time()
        elapsed_time = current_time - self.time_start
        elapsed_ftime = strftime("%H:%M:%S", gmtime(elapsed_time))

        content = f"elapsed time: {elapsed_ftime}"
        content += f" | {message}"
        content = f"\n[{content}]\n"

        sys.stdout.write(content)
        sys.stdout.flush()



from abc import abstractmethod
from itertools import combinations

import numpy as np


class MacroLosses:
    def __call__(self, cp_mats):
        off_loss = [
            [np.delete(a, i, 0).sum(0) for i in range(self.n_classes)] for a in cp_mats
        ]
        obj = np.array(off_loss).flatten()
        return obj


class MicroLosses:
    def __call__(self, cp_mats):
        u = np.array(
            [[np.delete(a, i, 0) for i in range(self.n_classes)] for a in cp_mats]
        )
        p = np.array(
            [
                [np.delete(a, i).reshape(-1, 1) for i in range(self.n_classes)]
                for a in self.p_vecs
            ]
        )
        w = np.array(
            [
                [p[i, j] * u[i, j] for j in range(self.n_classes)]
                for i in range(self.n_groups)
            ]
        )
        obj = w.sum(2).flatten()
        return obj


class ConstraintBase:
    @abstractmethod
    def get_conditions(self):
        pass

    def _constraint_weights(self, p_vecs, p_a, cp_mats):
        # Getting the costraint weights
        constraints_by_group = [
            self.__get_constraints(p_vecs[i], p_a[i], cp_mats[i])
            for i in range(self.n_groups)
        ]

        # Arranging the constraint weights by group comparisons
        return self.__pair_constraints(constraints_by_group, cp_mats)

    def __get_constraints(self, p_vec, p_a, cp_mat):
        """Calculates TPR and FPR weights for the constraint matrix"""
        # Shortening the vars to keep things clean
        p = p_vec
        M = cp_mat

        # Setting up the matrix of parameter weights
        n_classes = M.shape[0]
        n_params = n_classes**2
        tpr = np.zeros(shape=(n_classes, n_params))
        fpr = np.zeros(shape=(n_classes, n_params))

        # Filling in the weights
        for i in range(n_classes):
            # Dropping row to calculate FPR
            p_i = np.delete(p, i)
            M_i = np.delete(M, i, 0)

            start = i * (n_classes)
            end = start + n_classes
            fpr[i, start:end] = np.dot(p_i, M_i) / p_i.sum()
            tpr[i, start:end] = M[i]

        # Reshaping the off-diagonal constraints
        strict = np.zeros(shape=(n_params, n_params))
        # A = np.array(p.T / p_a).T
        # B = np.array(M.T * A).T
        B = M
        for i in range(n_classes):
            start = i * n_classes
            end = start + n_classes
            strict[start:end, start:end] = B

        return tpr, fpr, strict

    def __pair_constraints(self, constraints, cp_mats):
        """Takes the output of constraint_weights() and returns a matrix
        of the pairwise constraints
        """
        # Setting up the preliminaries
        tprs = np.array([c[0] for c in constraints])
        fprs = np.array([c[1] for c in constraints])
        strict = np.array([c[2] for c in constraints])
        n_params = tprs.shape[2]
        n_classes = tprs.shape[1]
        n_groups = self.n_groups
        if n_groups > 2:
            group_combos = list(combinations(range(n_groups), 2))[:-1]
        else:
            group_combos = [(0, 1)]

        n_pairs = len(group_combos)

        # Setting up the empty matrices
        tpr_cons = np.zeros(shape=(n_pairs, n_groups, n_classes, n_params))
        fpr_cons = np.zeros(shape=(n_pairs, n_groups, n_classes, n_params))
        strict_cons = np.zeros(shape=(n_pairs, n_groups, n_params, n_params))

        # Filling in the constraint comparisons
        for i, c in enumerate(group_combos):
            # Getting the original diffs for flipping
            diffs = cp_mats[c[0]] - cp_mats[c[1]]
            tpr_flip = np.sign(np.diag(diffs)).reshape(-1, 1)
            fpr_flip = np.sign(np.sum(fprs[c[0]], 1) - np.sum(fprs[c[1]], 1))
            fpr_flip = fpr_flip.reshape(-1, 1)
            fpr_flip[np.where(fpr_flip == 0)] = 1

            # Filling in the constraints
            tpr_cons[i, c[0]] = tpr_flip * tprs[c[0]]
            tpr_cons[i, c[1]] = tpr_flip * -1 * tprs[c[1]]
            fpr_cons[i, c[0]] = fpr_flip * fprs[c[0]]
            fpr_cons[i, c[1]] = fpr_flip * -1 * fprs[c[1]]
            strict_cons[i, c[0]] = strict[c[0]]
            strict_cons[i, c[1]] = -1 * strict[c[1]]

        # Filling in the norm constraints
        one_cons = np.zeros(shape=(n_groups * n_classes, n_groups * n_params))
        cols = np.array(list(range(0, n_groups * n_classes**2, n_classes)))
        cols = cols.reshape(n_groups, n_classes)
        i = 0
        for c in cols:
            for j in range(n_classes):
                one_cons[i, c + j] = 1
                i += 1

        # Reshaping the arrays
        tpr_cons = np.concatenate([np.hstack(m) for m in tpr_cons])
        fpr_cons = np.concatenate([np.hstack(m) for m in fpr_cons])
        strict_cons = np.concatenate([np.hstack(m) for m in strict_cons])

        return tpr_cons, fpr_cons, strict_cons, one_cons


class EqualizedOdds(ConstraintBase):
    def get_conditions(self, p_vecs, p_a, cp_mats):
        tpr_cons, fpr_cons, strict_cons, norm_cons = self._constraint_weights(
            p_vecs, p_a, cp_mats
        )
        norm_bounds = np.repeat(1, norm_cons.shape[0])
        con = np.concatenate([tpr_cons, fpr_cons, norm_cons])
        eo_bounds = np.repeat(0, tpr_cons.shape[0] * 2)
        con_bounds = np.concatenate([eo_bounds, norm_bounds])
        return con, con_bounds


class EqualizedOpportunity(ConstraintBase):
    def get_conditions(self, p_vecs, p_a, cp_mats):
        tpr_cons, fpr_cons, strict_cons, norm_cons = self._constraint_weights(
            p_vecs, p_a, cp_mats
        )
        norm_bounds = np.repeat(1, norm_cons.shape[0])
        con = np.concatenate([tpr_cons, norm_cons])
        tpr_bounds = np.repeat(0, tpr_cons.shape[0])
        con_bounds = np.concatenate([tpr_bounds, norm_bounds])
        return con, con_bounds


class Strict(ConstraintBase):
    def get_conditions(self, p_vecs, p_a, cp_mats):
        tpr_cons, fpr_cons, strict_cons, norm_cons = self._constraint_weights(
            p_vecs, p_a, cp_mats
        )
        norm_bounds = np.repeat(1, norm_cons.shape[0])
        con = np.concatenate([strict_cons, norm_cons])
        strict_bounds = np.repeat(0, strict_cons.shape[0])
        con_bounds = np.concatenate([strict_bounds, norm_bounds])
        return con, con_bounds



from collections import defaultdict
from itertools import product

from ._categorical_feature import CategoricalFeature
from ._utils import (
    assign_overflow,
    flow_on_group_features,
    gen_desired_count,
    gen_desired_dist,
    get_categories_count,
    get_categories_count_norm,
    get_group_data,
    get_median,
    get_median_per_category,
    get_mode,
)


class CategoricalRepairer:
    def __init__(
        self, feature_to_repair, repair_level, kdd=False, features_to_ignore=[]
    ):
        self.feature_to_repair = feature_to_repair
        self.repair_level = repair_level
        self.kdd = kdd
        self.features_to_ignore = features_to_ignore

    def repair(self, data_to_repair):
        num_cols = len(data_to_repair[0])
        col_ids = list(range(num_cols))

        # Get column type information
        col_types = ["Y"] * len(col_ids)
        for i, col in enumerate(col_ids):
            if i in self.features_to_ignore:
                col_types[i] = "I"
            elif i == self.feature_to_repair:
                col_types[i] = "X"

        col_type_dict = {
            col_id: col_type for col_id, col_type in zip(col_ids, col_types)
        }

        not_I_col_ids = [x for x in col_ids if col_type_dict[x] != "I"]

        if self.kdd:
            cols_to_repair = [x for x in col_ids if col_type_dict[x] == "Y"]
        else:
            cols_to_repair = [x for x in col_ids if col_type_dict[x] in "YX"]

        # To prevent potential perils with user-provided column names, map them to safe column names
        safe_stratify_cols = [self.feature_to_repair]

        # Extract column values for each attribute in data
        data_dict = {col_id: [] for col_id in col_ids}

        # Populate each attribute with its column values
        for row in data_to_repair:
            for i in col_ids:
                data_dict[i].append(row[i])

        repair_types = {}
        for col_id, values in list(data_dict.items()):
            if all(isinstance(value, float) for value in values):
                repair_types[col_id] = float
            elif all(isinstance(value, int) for value in values):
                repair_types[col_id] = int
            else:
                repair_types[col_id] = str

        unique_col_vals = {}
        index_lookup = {}
        for col_id in not_I_col_ids:
            col_values = data_dict[col_id]
            # extract unique values from column and sort
            col_values = sorted(list(set(col_values)))
            unique_col_vals[col_id] = col_values
            # look up a value, get its position
            index_lookup[col_id] = {col_values[i]: i for i in range(len(col_values))}

        unique_stratify_values = [unique_col_vals[i] for i in safe_stratify_cols]
        all_stratified_groups = list(product(*unique_stratify_values))
        # look up a stratified group, and get a list of indices corresponding to that group in the data
        stratified_group_indices = defaultdict(list)

        # Find the number of unique values for each strat-group, organized per column.
        val_sets = {
            group: {col_id: set() for col_id in cols_to_repair}
            for group in all_stratified_groups
        }
        for i, row in enumerate(data_to_repair):
            group = tuple(row[col] for col in safe_stratify_cols)
            for col_id in cols_to_repair:
                val_sets[group][col_id].add(row[col_id])

            # Also remember that this row pertains to this strat-group.
            stratified_group_indices[group].append(i)

        stratified_group_data = {group: {} for group in all_stratified_groups}
        for group in all_stratified_groups:
            for col_id, col_dict in list(data_dict.items()):
                # Get the indices at which each value occurs.
                indices = {}
                for i in stratified_group_indices[group]:
                    value = col_dict[i]
                    if value not in indices:
                        indices[value] = []
                    indices[value].append(i)

                stratified_col_values = [
                    (occurs, val) for val, occurs in list(indices.items())
                ]
                stratified_col_values.sort(key=lambda tup: tup[1])
                stratified_group_data[group][col_id] = stratified_col_values

        mode_feature_to_repair = get_mode(data_dict[self.feature_to_repair])

        # Repair Data and retrieve the results
        for col_id in cols_to_repair:
            # which bucket value we're repairing
            group_offsets = {group: 0 for group in all_stratified_groups}
            col = data_dict[col_id]

            num_quantiles = min(
                len(val_sets[group][col_id]) for group in all_stratified_groups
            )
            quantile_unit = 1.0 / num_quantiles

            if repair_types[col_id] in {int, float}:
                for quantile in range(num_quantiles):
                    median_at_quantiles = []
                    indices_per_group = {}

                    for group in all_stratified_groups:
                        group_data_at_col = stratified_group_data[group][col_id]
                        num_vals = len(group_data_at_col)
                        offset = int(round(group_offsets[group] * num_vals))
                        number_to_get = int(
                            round((group_offsets[group] + quantile_unit) * num_vals)
                            - offset
                        )
                        group_offsets[group] += quantile_unit

                        if number_to_get > 0:

                            # Get data at this quantile from this Y column such that stratified X = group
                            offset_data = group_data_at_col[
                                offset : offset + number_to_get
                            ]
                            indices_per_group[group] = [
                                i for val_indices, _ in offset_data for i in val_indices
                            ]
                            values = sorted([float(val) for _, val in offset_data])

                            # Find this group's median value at this quantile
                            median_at_quantiles.append(get_median(values, self.kdd))

                    # Find the median value of all groups at this quantile (chosen from each group's medians)
                    median = get_median(median_at_quantiles, self.kdd)
                    median_val_pos = index_lookup[col_id][median]

                    # Update values to repair the dataset.
                    for group in all_stratified_groups:
                        for index in indices_per_group[group]:
                            original_value = col[index]

                            current_val_pos = index_lookup[col_id][original_value]
                            distance = (
                                median_val_pos - current_val_pos
                            )  # distance between indices
                            distance_to_repair = int(
                                round(distance * self.repair_level)
                            )
                            index_of_repair_value = current_val_pos + distance_to_repair
                            repaired_value = unique_col_vals[col_id][
                                index_of_repair_value
                            ]

                            # Update data to repaired valued
                            data_dict[col_id][index] = repaired_value

            # Categorical Repair is done below
            elif repair_types[col_id] in {str}:
                feature = CategoricalFeature(col)
                categories = list(feature.bin_index_dict.keys())

                group_features = get_group_data(
                    all_stratified_groups, stratified_group_data, col_id
                )

                categories_count = get_categories_count(
                    categories, all_stratified_groups, group_features
                )

                categories_count_norm = get_categories_count_norm(
                    categories, all_stratified_groups, categories_count, group_features
                )

                median = get_median_per_category(categories, categories_count_norm)

                # Partially fill-out the generator functions to simplify later calls.
                dist_generator = lambda group_index, category: gen_desired_dist(
                    group_index,
                    category,
                    col_id,
                    median,
                    self.repair_level,
                    categories_count_norm,
                    self.feature_to_repair,
                    mode_feature_to_repair,
                )

                count_generator = (
                    lambda group_index, group, category: gen_desired_count(
                        group_index,
                        group,
                        category,
                        median,
                        group_features,
                        self.repair_level,
                        categories_count,
                    )
                )

                group_features, overflow = flow_on_group_features(
                    all_stratified_groups, group_features, count_generator
                )

                group_features, assigned_overflow, distribution = assign_overflow(
                    all_stratified_groups,
                    categories,
                    overflow,
                    group_features,
                    dist_generator,
                )

                # Return our repaired feature in the form of our original dataset
                for group in all_stratified_groups:
                    indices = stratified_group_indices[group]
                    for i, index in enumerate(indices):
                        repaired_value = group_features[group].data[i]
                        data_dict[col_id][index] = repaired_value

        # Replace stratified groups with their mode value, to remove it's information
        repaired_data = []
        for i, orig_row in enumerate(data_to_repair):
            new_row = [
                orig_row[j] if j not in cols_to_repair else data_dict[j][i]
                for j in col_ids
            ]
            repaired_data.append(new_row)
        return repaired_data



import sys
import time

import numpy as np
from tqdm import tqdm


class GAHiperparameters(object):
    def __init__(self, **kargs):
        self.max_num_iteration = kargs.get("max_num_iteration", None)
        self.population_size = kargs.get("population_size", 100)
        self.mutation_probability = kargs.get("mutation_probability", 0.1)
        self.elit_ratio = kargs.get("elit_ratio", 0.01)
        self.crossover_probability = kargs.get("crossover_probability", 0.5)
        self.parents_portion = kargs.get("parents_portion", 0.3)
        self.crossover_type = kargs.get("crossover_type", "uniform")
        self.max_iteration_without_improv = kargs.get(
            "max_iteration_without_improv", None
        )


class GeneticAlgorithm(object):
    """
    Genetic Algorithm (Elitist version)
    Implementation of elitist genetic algorithm for solving problems with
    continuous, integers, or mixed variables.
    """

    def __init__(
        self,
        function,
        dimension: int,
        variable_type: str = "bool",
        variable_boundaries: np.ndarray = None,
        variable_type_mixed: np.ndarray = None,
        algorithm_parameters: GAHiperparameters = None,
        verbose: int = 0,
    ):
        """
        Parameters
        ----------

        function : <Callable>
        The given objective function to be minimized.

        dimension : int
        The number of decision variables.

        variable_type: str
            - 'bool' if all variables are Boolean;
            - 'int' if all variables are integer; and
            - 'real' if all variables are real value or continuous

        for mixed type see @param variable_type_mixed.

        variable_boundaries: array-like
        None if variable_type is 'bool'; otherwise provide an array of tuples
        of length two as boundaries for each variable;
        the length of the array must be equal dimension. For example,
        np.array([0,100],[0,200]) determines lower boundary 0 and upper boundary 100 for first
        and upper boundary 200 for second variable where dimension is 2.

        variable_type_mixed: array-like
        None if all variables have the same type; otherwise this can be used to
        specify the type of each variable separately. For example if the first
        variable is integer but the second one is real the input is:
        np.array(['int'],['real']). NOTE: it does not accept 'bool'. If variable
        type is Boolean use 'int' and provide a boundary as [0,1]
        in variable_boundaries. Also if variable_type_mixed is applied,
        variable_boundaries has to be defined.

        algorithm_parameters: GAArguments
        class containing GA algoirhtm hiperparameters

        verbose: int
        If >0, log information

        """
        self.verbose = verbose

        # input function
        assert callable(function), "function must be callable"
        self.f = function
        self.dim = int(dimension)

        # input variable type
        assert (
            variable_type == "bool" or variable_type == "int" or variable_type == "real"
        ), "\n variable_type must be 'bool', 'int', or 'real'"

        # input variables' type (MIXED)
        if variable_type_mixed is None:

            if variable_type == "real":
                self.var_type = np.array([["real"]] * self.dim)
            else:
                self.var_type = np.array([["int"]] * self.dim)
        else:
            assert (
                type(variable_type_mixed).__module__ == "numpy"
            ), "\n variable_type must be numpy array"
            assert (
                len(variable_type_mixed) == self.dim
            ), "\n variable_type must have a length equal dimension."

            for i in variable_type_mixed:
                assert i == "real" or i == "int", (
                    "\n variable_type_mixed is either 'int' or 'real' "
                    + "ex:['int','real','real']"
                    + "\n for 'boolean' use 'int' and specify boundary as [0,1]"
                )

            self.var_type = variable_type_mixed

        # input variables' boundaries
        if variable_type != "bool" or type(variable_type_mixed).__module__ == "numpy":

            assert (
                type(variable_boundaries).__module__ == "numpy"
            ), "\n variable_boundaries must be numpy array"

            assert (
                len(variable_boundaries) == self.dim
            ), "\n variable_boundaries must have a length equal dimension"

            for i in variable_boundaries:
                assert (
                    len(i) == 2
                ), "\n boundary for each variable must be a tuple of length two."
                assert (
                    i[0] <= i[1]
                ), "\n lower_boundaries must be smaller than upper_boundaries [lower,upper]"
            self.var_bound = variable_boundaries
        else:
            self.var_bound = np.array([[0, 1]] * self.dim)

        # input algorithm's parameters
        self.param = algorithm_parameters.__dict__

        self.pop_s = int(self.param["population_size"])

        assert (
            self.param["parents_portion"] <= 1 and self.param["parents_portion"] >= 0
        ), "parents_portion must be in range [0,1]"

        self.par_s = int(self.param["parents_portion"] * self.pop_s)
        trl = self.pop_s - self.par_s
        if trl % 2 != 0:
            self.par_s += 1

        self.prob_mut = self.param["mutation_probability"]

        assert (
            self.prob_mut <= 1 and self.prob_mut >= 0
        ), "mutation_probability must be in range [0,1]"

        self.prob_cross = self.param["crossover_probability"]
        assert (
            self.prob_cross <= 1 and self.prob_cross >= 0
        ), "mutation_probability must be in range [0,1]"

        assert (
            self.param["elit_ratio"] <= 1 and self.param["elit_ratio"] >= 0
        ), "elit_ratio must be in range [0,1]"

        trl = self.pop_s * self.param["elit_ratio"]
        if trl < 1 and self.param["elit_ratio"] > 0:
            self.num_elit = 1
        else:
            self.num_elit = int(trl)

        assert (
            self.par_s >= self.num_elit
        ), "\n number of parents must be greater than number of elits"

        if self.param["max_num_iteration"] == None:
            self.iterate = 0
            for i in range(0, self.dim):
                if self.var_type[i] == "int":
                    self.iterate += (
                        (self.var_bound[i][1] - self.var_bound[i][0])
                        * self.dim
                        * (100 / self.pop_s)
                    )
                else:
                    self.iterate += (
                        (self.var_bound[i][1] - self.var_bound[i][0])
                        * 50
                        * (100 / self.pop_s)
                    )
            self.iterate = int(self.iterate)
            if (self.iterate * self.pop_s) > 10000000:
                self.iterate = 10000000 / self.pop_s
        else:
            self.iterate = int(self.param["max_num_iteration"])

        self.c_type = self.param["crossover_type"]
        assert (
            self.c_type == "uniform"
            or self.c_type == "one_point"
            or self.c_type == "two_point"
        ), "\n crossover_type must 'uniform', 'one_point', or 'two_point' Enter string"

        self.stop_mniwi = False
        if self.param["max_iteration_without_improv"] == None:
            self.mniwi = self.iterate + 1
        else:
            self.mniwi = int(self.param["max_iteration_without_improv"])

    def run(self):

        # Initial Population
        self.integers = np.where(self.var_type == "int")
        self.reals = np.where(self.var_type == "real")

        pop = np.array([np.zeros(self.dim + 1)] * self.pop_s)
        solo = np.zeros(self.dim + 1)
        var = np.zeros(self.dim)

        for p in range(0, self.pop_s):

            for i in self.integers[0]:
                var[i] = np.random.randint(
                    self.var_bound[i][0], self.var_bound[i][1] + 1
                )
                solo[i] = var[i].copy()
            for i in self.reals[0]:
                var[i] = self.var_bound[i][0] + np.random.random() * (
                    self.var_bound[i][1] - self.var_bound[i][0]
                )
                solo[i] = var[i].copy()

            obj = self.sim(var)
            solo[self.dim] = obj
            pop[p] = solo.copy()

        # Report
        self.report = []
        self.test_obj = obj
        self.best_variable = var.copy()
        self.best_function = obj

        t = 1
        counter = 0
        with tqdm(total=self.iterate) as pbar:
            while t <= self.iterate:

                if self.verbose == True:
                    pbar.update(1)

                # Sort
                pop = pop[pop[:, self.dim].argsort()]

                if pop[0, self.dim] < self.best_function:
                    counter = 0
                    self.best_function = pop[0, self.dim].copy()
                    self.best_variable = pop[0, : self.dim].copy()
                    pbar.set_description(f"Cost: {self.best_function:.4f}")
                    pbar.refresh()
                else:
                    counter += 1
                # Report
                self.report.append(pop[0, self.dim])

                # Normalizing objective function
                normobj = np.zeros(self.pop_s)

                minobj = pop[0, self.dim]
                if minobj < 0:
                    normobj = pop[:, self.dim] + abs(minobj)

                else:
                    normobj = pop[:, self.dim].copy()

                maxnorm = np.amax(normobj)
                normobj = maxnorm - normobj + 1

                # Calculate probability
                sum_normobj = np.sum(normobj)
                prob = np.zeros(self.pop_s)
                prob = normobj / sum_normobj
                cumprob = np.cumsum(prob)

                # Select parents
                par = np.array([np.zeros(self.dim + 1)] * self.par_s)

                for k in range(0, self.num_elit):
                    par[k] = pop[k].copy()
                for k in range(self.num_elit, self.par_s):
                    index = np.searchsorted(cumprob, np.random.random())
                    par[k] = pop[index].copy()

                ef_par_list = np.array([False] * self.par_s)
                par_count = 0
                while par_count == 0:
                    for k in range(0, self.par_s):
                        if np.random.random() <= self.prob_cross:
                            ef_par_list[k] = True
                            par_count += 1

                ef_par = par[ef_par_list].copy()

                # New generation
                pop = np.array([np.zeros(self.dim + 1)] * self.pop_s)

                for k in range(0, self.par_s):
                    pop[k] = par[k].copy()

                for k in range(self.par_s, self.pop_s, 2):
                    r1 = np.random.randint(0, par_count)
                    r2 = np.random.randint(0, par_count)
                    pvar1 = ef_par[r1, : self.dim].copy()
                    pvar2 = ef_par[r2, : self.dim].copy()

                    ch = self.cross(pvar1, pvar2, self.c_type)
                    ch1 = ch[0].copy()
                    ch2 = ch[1].copy()

                    ch1 = self.mut(ch1)
                    ch2 = self.mutmidle(ch2, pvar1, pvar2)
                    solo[: self.dim] = ch1.copy()
                    obj = self.sim(ch1)
                    solo[self.dim] = obj
                    pop[k] = solo.copy()
                    solo[: self.dim] = ch2.copy()
                    obj = self.sim(ch2)
                    solo[self.dim] = obj
                    pop[k + 1] = solo.copy()

                t += 1
                if counter > self.mniwi:
                    pop = pop[pop[:, self.dim].argsort()]
                    if pop[0, self.dim] >= self.best_function:
                        t = self.iterate
                        if self.verbose == True:
                            pbar.update(t - pbar.n)
                        time.sleep(2)
                        t += 1
                        self.stop_mniwi = True

        # Sort
        pop = pop[pop[:, self.dim].argsort()]

        if pop[0, self.dim] < self.best_function:

            self.best_function = pop[0, self.dim].copy()
            self.best_variable = pop[0, : self.dim].copy()
            pbar.set_description(f"Cost: {self.best_function:.4f}")
            pbar.refresh()

        # Report
        self.report.append(pop[0, self.dim])

        self.output_dict = {
            "variable": self.best_variable,
            "function": self.best_function,
        }
        if self.verbose == True:
            show = " " * 100
            sys.stdout.write("\r%s" % (show))
            sys.stdout.write("\r The best solution found:\n %s" % (self.best_variable))
            sys.stdout.write("\n\n Objective function:\n %s\n" % (self.best_function))
            sys.stdout.flush()
        re = np.array(self.report)

        if self.stop_mniwi == True:
            sys.stdout.write(
                "\nWarning: GA is terminated due to the"
                + " maximum number of iterations without improvement was met!"
            )

    def cross(self, x, y, c_type):

        ofs1 = x.copy()
        ofs2 = y.copy()

        if c_type == "one_point":
            ran = np.random.randint(0, self.dim)
            for i in range(0, ran):
                ofs1[i] = y[i].copy()
                ofs2[i] = x[i].copy()

        if c_type == "two_point":

            ran1 = np.random.randint(0, self.dim)
            ran2 = np.random.randint(ran1, self.dim)

            for i in range(ran1, ran2):
                ofs1[i] = y[i].copy()
                ofs2[i] = x[i].copy()

        if c_type == "uniform":

            for i in range(0, self.dim):
                ran = np.random.random()
                if ran < 0.5:
                    ofs1[i] = y[i].copy()
                    ofs2[i] = x[i].copy()

        return np.array([ofs1, ofs2])

    def mut(self, x):

        for i in self.integers[0]:
            ran = np.random.random()
            if ran < self.prob_mut:

                x[i] = np.random.randint(self.var_bound[i][0], self.var_bound[i][1] + 1)

        for i in self.reals[0]:
            ran = np.random.random()
            if ran < self.prob_mut:

                x[i] = self.var_bound[i][0] + np.random.random() * (
                    self.var_bound[i][1] - self.var_bound[i][0]
                )

        return x

    def mutmidle(self, x, p1, p2):
        for i in self.integers[0]:
            ran = np.random.random()
            if ran < self.prob_mut:
                if p1[i] < p2[i]:
                    x[i] = np.random.randint(p1[i], p2[i])
                elif p1[i] > p2[i]:
                    x[i] = np.random.randint(p2[i], p1[i])
                else:
                    x[i] = np.random.randint(
                        self.var_bound[i][0], self.var_bound[i][1] + 1
                    )

        for i in self.reals[0]:
            ran = np.random.random()
            if ran < self.prob_mut:
                if p1[i] < p2[i]:
                    x[i] = p1[i] + np.random.random() * (p2[i] - p1[i])
                elif p1[i] > p2[i]:
                    x[i] = p2[i] + np.random.random() * (p1[i] - p2[i])
                else:
                    x[i] = self.var_bound[i][0] + np.random.random() * (
                        self.var_bound[i][1] - self.var_bound[i][0]
                    )
        return x

    def evaluate(self):
        return self.f(self.temp)

    def sim(self, X):
        self.temp = X.copy()
        obj = self.evaluate()
        return obj



"""
The :mod:`holisticai.bias.mitigation` module includes preprocessing, inprocessing and postprocessing bias mitigation algorithms.
"""

# inprocessing algorithm classes
from .inprocessing import (
    BlindSpotAwareMF,
    DebiasingLearningMF,
    ExponentiatedGradientReduction,
    FairKCenterClustering,
    FairKmedianClustering,
    FairletClustering,
    FairRec,
    FairScoreClassifier,
    GridSearchReduction,
    MetaFairClassifier,
    PopularityPropensityMF,
    PrejudiceRemover,
    VariationalFairClustering,
)

# postprocessing algorithm classes
from .postprocessing import (
    CalibratedEqualizedOdds,
    DebiasingExposure,
    EqualizedOdds,
    FairTopK,
    LPDebiaserBinary,
    LPDebiaserMulticlass,
    MLDebiaser,
    PluginEstimationAndCalibration,
    RejectOptionClassification,
    WassersteinBarycenter,
)

# preprocessing algorithm classes
from .preprocessing import (
    CorrelationRemover,
    FairletClusteringPreprocessing,
    LearningFairRepresentation,
    Reweighing,
)

# all
__all__ = [
    "CorrelationRemover",
    "Reweighing",
    "LearningFairRepresentation",
    "ExponentiatedGradientReduction",
    "GridSearchReduction",
    "CalibratedEqualizedOdds",
    "EqualizedOdds",
    "RejectOptionClassification",
    "WassersteinBarycenter",
    "MLDebiaser",
    "LPDebiaserBinary",
    "LPDebiaserMulticlass",
    "PluginEstimationAndCalibration",
    "PrejudiceRemover",
    "MetaFairClassifier",
    "VariationalFairClustering",
    "FairletClustering",
    "FairletClusteringPreprocessing",
    "FairKCenterClustering",
    "FairKmedianClustering",
    "BlindSpotAwareMF",
    "DebiasingLearningMF",
    "PopularityPropensityMF",
    "FairRec",
    "FairScoreClassifier",
    "DebiasingExposure",
    "FairTopK",
]

import importlib

torch_spec = importlib.util.find_spec("torch")
if torch_spec is not None:
    from .inprocessing import AdversarialDebiasing
__all__ += ["AdversarialDebiasing"]

networkx_spec = importlib.util.find_spec("networkx")
if networkx_spec is not None:
    from .postprocessing import DisparateImpactRemoverRS
    from .preprocessing import DisparateImpactRemover

__all__ += ["DisparateImpactRemoverRS", "DisparateImpactRemover"]



import warnings

import numpy as np


def check_ranking(protected, mtable):
    """
    Description
    -----------
    Checks if the ranking is fair in respect to the mtable

    Parameters
    ----------
    ranking: DataFrame
        The ranking to be checked.

    mtable: List
        The mtable against to check (list of int)

    Return
    ------
        bool
        Returns whether the rankings satisfies the mtable
    """
    # if the mtable has a different number elements than there are in the top docs return false
    if len(protected) != len(mtable):
        raise ValueError(
            "Number of documents in ranking and mtable length must be equal!"
        )

    # check number of protected element at each rank
    return not (protected.cumsum().values < np.array(mtable)).any()


def validate_basic_parameters(k, p, alpha):
    """
    Description
    -----------
    Validates if k, p and alpha are in the required ranges

    Parameters
    ----------
        k : int
            Total number of elements (above or equal to 10)

        p : int
            The proportion of protected candidates in the top-k ranking (between 0.02 and 0.98)

        alpha : float
            The significance level (between 0.01 and 0.15)
    """
    if k < 10 or k > 400:
        if k < 2:
            raise ValueError(
                "Total number of elements `k` should be between 10 and 400"
            )
        else:
            warnings.warn("Library has not been tested with values outside this range")

    if p < 0.02 or p > 0.98:
        if p < 0 or p > 1:
            raise ValueError(
                "The proportion of protected candidates `p` in the top-k ranking should be between "
                "0.02 and 0.98"
            )
        else:
            warnings.warn("Library has not been tested with values outside this range")

    validate_alpha(alpha)


def validate_alpha(alpha):
    if alpha < 0.01 or alpha > 0.15:
        if alpha < 0.001 or alpha > 0.5:
            raise ValueError(
                "The significance level `alpha` must be between 0.01 and 0.15"
            )
        else:
            warnings.warn("Library has not been tested with values outside this range")



import sys
from typing import Optional

import numpy as np
import scipy.optimize as optim
from scipy.spatial.distance import cdist
from scipy.special import softmax

from holisticai.utils.transformers.bias import BMPreprocessing


def get_xhat_y_hat(prototypes, w, x):
    M = softmax(-cdist(x, prototypes), axis=1)
    x_hat = np.matmul(M, prototypes)
    y_hat = np.clip(
        np.matmul(M, w.reshape((-1, 1))), np.finfo(float).eps, 1.0 - np.finfo(float).eps
    )
    return M, x_hat, y_hat


class ObjectiveFunction:
    def __init__(self, verbose, print_interval):
        self.verbose = verbose
        self.step = 0
        self.print_interval = print_interval

    def log_progress(self, total_loss, L_x, L_y, L_z):
        self.step += 1
        if self.verbose and (self.step % self.print_interval) == 0:
            sys.stdout.write(
                f"step: {self.step}\tloss: {total_loss:.4f}\tL_x: {L_x:.4f}\tL_y: {L_y:.4f}\tL_z: {L_z:.4f}\n"
            )
            sys.stdout.flush()

    def __call__(
        self, parameters, x_a, x_b, y_a, y_b, k=10, A_x=0.01, A_y=0.1, A_z=0.5
    ):

        features_dim = x_a.shape[1]

        w = parameters[:k]
        prototypes = parameters[k:].reshape((k, features_dim))

        M_b, x_hat_b, y_hat_b = get_xhat_y_hat(prototypes, w, x_b)
        M_a, x_hat_a, y_hat_a = get_xhat_y_hat(prototypes, w, x_a)

        y_hat = np.concatenate([y_hat_b, y_hat_a], axis=0)
        y = np.concatenate([y_b.reshape((-1, 1)), y_a.reshape((-1, 1))], axis=0)

        L_x = np.mean((x_hat_b - x_b) ** 2) + np.mean((x_hat_a - x_a) ** 2)
        L_z = np.mean(abs(np.mean(M_b, axis=0) - np.mean(M_a, axis=0)))
        L_y = -np.mean(y * np.log(y_hat) + (1.0 - y) * np.log(1.0 - y_hat))

        total_loss = A_x * L_x + A_y * L_y + A_z * L_z

        self.log_progress(total_loss, L_x, L_y, L_z)

        return total_loss


class LearningFairRepresentation(BMPreprocessing):
    """
    Learning fair representations finds a latent representation which encodes the data well
    while obfuscates information about protected attributes.

    References
    ----------
        Zemel, Rich, et al. "Learning fair representations."
        International conference on machine learning. PMLR, 2013.
    """

    def __init__(
        self,
        k: Optional[int] = 5,
        Ax: Optional[float] = 0.01,
        Ay: Optional[float] = 1.0,
        Az: Optional[float] = 50.0,
        print_interval: Optional[int] = 250,
        maxiter: Optional[int] = 5000,
        maxfun: Optional[int] = 5000,
        verbose: Optional[int] = 0,
        seed: Optional[int] = None,
    ):
        """
        Learning Fair Representation Preprocessing Bias Mitigator

        Description
        -----------
        Initialize Mitigator class.

        Parameters
        ----------
        k : int
            Number of prototypes.
        Ax : float
            Input recontruction quality term weight.
        Az : float
            Fairness constraint term weight.
        Ay : float
            Output prediction error.
        print_interval : int
            Print optimization objective value every print_interval iterations.
        verbose : int
            If zero, then no output.
        seed : int
            Seed to make `predict` repeatable.
        """
        self.seed = seed
        self.k = k
        self.Ax = Ax
        self.Ay = Ay
        self.Az = Az

        self.w = None
        self.prototypes = None
        self.learned_model = None
        self.maxiter = maxiter
        self.maxfun = maxfun
        self.obj_fun = ObjectiveFunction(verbose=verbose, print_interval=print_interval)
        self.problem_type = "binary"

    def fit(
        self,
        X: np.ndarray,
        y_true: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
    ):
        """
        Fit.

        Description
        -----------
        Fit data to learn a fair representation transform.

        Parameters
        ----------
        X : matrix-like
            Input data
        y_true : array-like
            Target vector
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)

        Return
        ------
            Self
        """
        if self.seed is not None:
            np.random.seed(self.seed)

        params = self._load_data(
            X=X,
            y_true=y_true,
            group_a=group_a,
            group_b=group_b,
        )
        y_true = params["y_true"]
        group_a = params["group_a"]
        group_b = params["group_b"]
        X = params["X"]

        _, self.features_dim = np.shape(X)
        parameters_initialization = np.random.uniform(
            size=self.k + self.features_dim * self.k
        )
        parameters_bounds = [(0, 1)] * self.k + [
            (None, None)
        ] * self.features_dim * self.k
        args = (
            X[group_a == 1],
            X[group_b == 1],
            y_true[group_a == 1],
            y_true[group_b == 1],
            self.k,
            self.Ax,
            self.Ay,
            self.Az,
        )

        self.learned_model = optim.fmin_l_bfgs_b(
            self.obj_fun,
            x0=parameters_initialization,
            epsilon=1e-5,
            args=args,
            bounds=parameters_bounds,
            approx_grad=True,
            maxfun=self.maxfun,
            maxiter=self.maxiter,
            disp=0,
        )[0]

        self.w = self.learned_model[: self.k]
        self.prototypes = self.learned_model[self.k :].reshape(
            (self.k, self.features_dim)
        )
        return self

    def transform(self, X: np.ndarray, group_a: np.ndarray, group_b: np.ndarray):
        """
        Transform data

        Description
        -----------
        Transform data to a fair representation

        Parameters
        ----------
        X : matrix-like
            Input data
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)

        Return
        ------
            Transformed X
        """
        params = self._load_data(X=X, group_a=group_a, group_b=group_b)
        X = params["X"]
        group_a = params["group_a"]
        group_b = params["group_b"]

        _, x_hat_b, _ = get_xhat_y_hat(self.prototypes, self.w, X[group_b == 1])
        _, x_hat_a, _ = get_xhat_y_hat(self.prototypes, self.w, X[group_a == 1])

        new_X = X.copy()
        new_X[group_b == 1] = x_hat_b
        new_X[group_a == 1] = x_hat_a

        return new_X

    def fit_transform(
        self,
        X: np.ndarray,
        y_true: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
    ):
        """
        Fit and transform

        Description
        ----------
        Fit and transform

        Parameters
        ----------
        X : matrix-like
            Input data
        y_true : array-like
            Target vector
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)

        Return
        ------
            Self
        """
        return self.fit(X, y_true, group_a, group_b).transform(X, group_a, group_b)



from typing import Optional

import numpy as np
import pandas as pd

from holisticai.utils.transformers.bias import BMPostprocessing as BMPost

from .algorithm_utils.fail_prob import RecursiveNumericFailProbabilityCalculator
from .algorithm_utils.valitation_utils import check_ranking, validate_basic_parameters


class FairTopK(BMPost):
    """
    Fair Top K bias mitigation can be used for Recommender Systems.
    The strategy extends group fairness definition using the standard notion of protected groups
    and is based on ensuring that the proportion of protected candidates in every prefix of the top-k
    ranking.

    Reference
    ---------
    Zehlike, Meike, et al. "Fa* ir: A fair top-k ranking algorithm." Proceedings of the 2017 ACM on
    Conference on Information and Knowledge Management. 2017.
    """

    def __init__(
        self,
        top_n: Optional[int],
        p: Optional[float],
        alpha: Optional[float],
        query_col: Optional[str] = "query_id",
        doc_col: Optional[str] = "doc_id",
        group_col: Optional[str] = "group_id",
        score_col: Optional[str] = "score",
    ):

        # check the parameters first
        validate_basic_parameters(top_n, p, alpha)
        self.query_col = query_col
        self.doc_col = doc_col
        self.group_col = group_col
        self.score_col = score_col

        # assign the parameters
        self.top_n = top_n  # the total number of elements
        self.p = p  # the proportion of protected candidates in the top-k ranking
        self.alpha = alpha  # the significance level
        self._cache = {}  # stores generated mtables in memory

    def transform(self, rankings, p_attr=None):
        """
        Apply transform to prediction scores.

        Parameters
        ----------
        rankings : DataFrame
            Predicted matrix scores (nb_examlpes*top_n, 3) [query_id, doc_id, scores]

        p_attr: matrix-like
            Item groups (nb_examples, 3) [query_id, doc_id, protected]
        """
        if p_attr is None:
            if not self.group_col in rankings.columns:
                raise ValueError("protected groups must be provided")
            else:
                new_rankings = rankings
        else:
            if self.group_col in rankings.columns:
                del rankings[self.group_col]

            new_rankings = pd.merge(
                rankings, p_attr, on=[self.query_col, self.doc_col], how="left"
            )

        query_result_by_group = new_rankings.groupby(self.query_col)
        re_rankings = [
            df if self.is_fair(df) else self.transform_ranking(df)
            for _, df in query_result_by_group
        ]
        return pd.concat(re_rankings).reset_index(drop=True)

    def transform_ranking(self, ranking):
        """
        Description
        -----------
        Applies FA*IR re-ranking to the input ranking using an adjusted mtable

        Parameters
        ----------
        ranking: list
            The ranking to be re-ranked (list of FairScoreDoc)

        Return
        ------
        :return:
        """
        protected = ranking[ranking[self.group_col] == True]
        non_protected = ranking[ranking[self.group_col] == False]
        mtable = self._create_adjusted_mtable()
        return pd.DataFrame(
            self._fair_top_k(protected, non_protected, mtable)
        ).reset_index(drop=True)

    def _create_adjusted_mtable(self):
        """
        Description
        -----------
        Creates an adjusted mtable by using the alpha value.

        Return
        ------
            list
            mtable as list of int elements
        """

        if not (self.top_n, self.p, self.alpha) in self._cache:
            # create the mtable
            fail_prob_pair = RecursiveNumericFailProbabilityCalculator(
                self.top_n, self.p, self.alpha
            ).adjust_alpha()
            mtable = [int(i) for i in fail_prob_pair.mtable.m.tolist()]
            # store as list
            self._cache[(self.top_n, self.p, self.alpha)] = mtable

        # return from cache
        return self._cache[(self.top_n, self.p, self.alpha)]

    def is_fair(self, ranking):
        """
        Description
        -----------
        Checks if the ranking is fair for the given parameters

        Parameters
        ----------
        ranking: list
            The ranking to be checked (list of Resultinfo)

        Return
        ------
            bool
        """
        return check_ranking(ranking[self.group_col], self._create_adjusted_mtable())

    def _fair_top_k(self, protected_candidates, non_protected_candidates, mtable):
        """
        Description
        -----------
        Reorganize the results info ensuring true the mtable condition (#protected[:i] >= mtable[i]).

        Parameters
        ----------

        protected_candidates:  pd.DataFrame
            ranking dataframe filtered with only protected candidates

        non_protected_candidates:  pd.DataFrame
            ranking dataframe filtered with only non protected candidates

        mtable: list
            adjusted mtable

        Return
        ------
            : list
            List of re-ranked results.
        """
        result = []
        countProtected = 0

        idxProtected = 0
        idxNonProtected = 0

        for i in range(self.top_n):
            if idxProtected >= len(protected_candidates) and idxNonProtected >= len(
                non_protected_candidates
            ):
                # no more candidates available, return list shorter than k
                return result
            if idxProtected >= len(protected_candidates):
                # no more protected candidates available, take non-protected instead
                result.append(non_protected_candidates.iloc[idxNonProtected])
                idxNonProtected += 1

            elif idxNonProtected >= len(non_protected_candidates):
                # no more non-protected candidates available, take protected instead
                result.append(protected_candidates.iloc[idxProtected])
                idxProtected += 1
                countProtected += 1
            elif countProtected < mtable[i]:
                # add a protected candidate
                result.append(protected_candidates.iloc[idxProtected])
                idxProtected += 1
                countProtected += 1
            else:
                # find the best candidate available
                if (
                    protected_candidates.iloc[idxProtected][self.score_col]
                    >= non_protected_candidates.iloc[idxNonProtected][self.score_col]
                ):
                    # the best is a protected one
                    result.append(protected_candidates.iloc[idxProtected])
                    idxProtected += 1
                    countProtected += 1
                else:
                    # the best is a non-protected one
                    result.append(non_protected_candidates.iloc[idxNonProtected])
                    idxNonProtected += 1

        return result



from copy import deepcopy

import numpy as np
from sklearn.metrics import accuracy_score, confusion_matrix


class CLFRates:
    def __init__(self, y_true, y_pred, round=4):
        self.tab = confusion_matrix(y_true, y_pred)
        tn = self.tab[0, 0]
        fn = self.tab[1, 0]
        fp = self.tab[0, 1]
        tp = self.tab[1, 1]
        self.pr = np.round((tp + fp) / len(y_true), round)
        self.nr = np.round((tn + fn) / len(y_true), round)
        self.tnr = np.round(tn / (tn + fp), round)
        self.tpr = np.round(tp / (tp + fn), round)
        self.fnr = np.round(fn / (fn + tp), round)
        self.fpr = np.round(fp / (fp + tn), round)
        self.acc = (tn + tp) / len(y_true)


def from_top(roc_point, round=4):
    d = np.sqrt(roc_point[0] ** 2 + (roc_point[1] - 1) ** 2)
    return d


def loss_from_roc(y, probs, roc):
    points = [(roc[0][i], roc[1][i]) for i in range(len(roc[0]))]
    guess_list = [threshold(probs, t) for t in roc[2]]
    accs = [accuracy_score(y, g) for g in guess_list]
    js = [p[1] - p[0] for p in points]
    tops = [from_top(point) for point in points]
    return {"guesses": guess_list, "accs": accs, "js": js, "tops": tops}


def pred_from_pya(y_pred, p_attr, pya, binom=False):
    # Getting the groups and making the initially all-zero predictor
    groups = np.unique(p_attr)
    out = deepcopy(y_pred)

    for i, g in enumerate(groups):
        group_ids = np.where((p_attr == g))[0]

        # Pulling the fitted switch probabilities for the group
        p = pya[i]

        # Indices in the group from which to choose swaps
        pos = np.where((p_attr == g) & (y_pred == 1))[0]
        neg = np.where((p_attr == g) & (y_pred == 0))[0]

        if not binom:
            # Randomly picking the positive predictions
            pos_samp = np.random.choice(a=pos, size=int(p[1] * len(pos)), replace=False)
            neg_samp = np.random.choice(a=neg, size=int(p[0] * len(neg)), replace=False)
            samp = np.concatenate((pos_samp, neg_samp)).flatten()
            out[samp] = 1
            out[np.setdiff1d(group_ids, samp)] = 0

    return out.astype(np.uint8)


# Quick function for thresholding probabilities
def threshold(probs, cutoff=0.5):
    return np.array(probs >= cutoff).astype(np.uint8)



from typing import Optional, Union

import numpy as np
from sklearn.base import BaseEstimator
from sklearn.metrics.pairwise import pairwise_distances_argmin

from holisticai.bias.mitigation.commons.fairlet_clustering.decompositions import (
    DecompositionMixin,
    ScalableFairletDecomposition,
    VanillaFairletDecomposition,
)
from holisticai.utils.models.cluster import KCenters, KMedoids
from holisticai.utils.transformers.bias import BMPreprocessing as BMPre

DECOMPOSITION_CATALOG = {
    "Scalable": ScalableFairletDecomposition,
    "Vanilla": VanillaFairletDecomposition,
}
CLUSTERING_CATALOG = {"KCenter": KCenters, "KMedoids": KMedoids}


class FairletClusteringPreprocessing(BaseEstimator, BMPre):
    """
    Fairlet decomposition is a pre-processing approach that computes
    fair micro-clusters where fairness is guaranteed. They then use
    the fairlet centers as a newly transformed dataset from the original.
    This transformed fairlet-based dataset is then provided to vanilla
    clustering algorithms, and hence, we obtain approximately
    fair clustering outputs as a result of the fairlets themselves being fair.

    References
    ----------
        Backurs, Arturs, et al. "Scalable fair clustering." International Conference on
        Machine Learning. PMLR, 2019.
    """

    def __init__(
        self,
        decomposition: Union["str", "DecompositionMixin"] = "Vanilla",
        p: Optional[str] = 1,
        q: Optional[float] = 3,
        seed: Optional[int] = None,
    ):
        """
        Parameters
        ----------
            decomposition : str
                Fairlet decomposition strategy, available: Vanilla, Scalable, MCF

            p : int
                fairlet decomposition parameter for Vanilla and Scalable strategy

            q : int
                fairlet decomposition parameter for Vanilla and Scalable strategy

            seed : int
                Random seed.
        """
        self.decomposition = DECOMPOSITION_CATALOG[decomposition](p=p, q=q)
        self.p = p
        self.q = q
        self.seed = seed

    def fit_transform(
        self,
        X: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
        sample_weight: Optional[np.ndarray] = None,
    ):
        """
        Fit the model

        Description
        -----------
        Learn a fair cluster.

        Parameters
        ----------

        X : matrix-like
            input matrix

        group_a : array-like
            binary mask vector

        group_b : array-like
            binary mask vector

        sample_weight (optional) : array-like
            Samples weights vector

        Returns
        -------
        the same object
        """
        params = self._load_data(
            X=X, sample_weight=sample_weight, group_a=group_a, group_b=group_b
        )
        X = params["X"]
        sample_weight = params["sample_weight"]
        group_a = params["group_a"].astype("int32")
        group_b = params["group_b"].astype("int32")
        np.random.seed(self.seed)
        fairlets, fairlet_centers, fairlet_costs = self.decomposition.fit_transform(
            X, group_a, group_b
        )
        Xt = np.zeros_like(X)
        mapping = np.zeros(len(X), dtype="int32")
        centers = np.array([X[fairlet_center] for fairlet_center in fairlet_centers])
        for i, fairlet in enumerate(fairlets):
            Xt[fairlet] = X[fairlet_centers[i]]
            mapping[fairlet] = i
            sample_weight[fairlet] = len(fairlet) / len(X)

        self.update_estimator_param("sample_weight", sample_weight)
        self.sample_weight = sample_weight
        self.X = X
        self.mapping = mapping
        self.centers = centers
        return Xt

    def transform(self, X):
        fairlets_midxs = pairwise_distances_argmin(X, Y=self.X)
        return self.centers[self.mapping[fairlets_midxs]]



import numpy as np
import pandas as pd

from .._conventions import _ALL, _LABEL
from .._moments_utils import BaseMoment


class ErrorRate(BaseMoment):
    """
    Extend BaseMoment for error rate objective.
    A classifier :math:`h(X)` has the misclassification error equal to
    .. math::
      P[h(X) \ne Y]
    """

    def load_data(self, X, y, sensitive_features):
        super().load_data(X, y, sensitive_features)
        self.index = [_ALL]

    def signed_weights(self):
        """Return the signed weights."""
        return 2 * self.tags[_LABEL] - 1

    def gamma(self, predictor):
        """Return the gamma values for the given predictor."""
        pred = predictor(self.X)

        if isinstance(pred, np.ndarray):
            pred = np.squeeze(pred)

        error = pd.Series(
            data=(self.tags[_LABEL] - pred).abs().mean(), index=self.index
        )
        return error



from typing import Optional

import numpy as np
from sklearn.decomposition import NMF

from holisticai.utils.models.recommender._rsbase import RecommenderSystemBase


class NonNegativeMF(RecommenderSystemBase):
    """
    This model is trained using conventional matrix factorization (Eq.1 [1] with Pu,i = 1),
    and the system always selects the top predicted item for each user, and adds
    it to the next (new) training set.

    References:
        [1] Sun, Wenlong, et al. "Debiasing the human-recommender system feedback loop in
        collaborative filtering." Companion Proceedings of The 2019 World Wide Web
        Conference. 2019.
    """

    def __init__(self, K: Optional[int] = 10):
        """
        Init Wrapper Non Negative Matrix Factorization

        Parameters
        ----------

        K : int
            Specifies the number of dimensions.
        """
        self.K = K

    def fit(self, X: Optional[np.ndarray]):
        """
        Fit model

        Parameters
        ----------

        X : matrix-like
            rating matrix, 0 means non-raked cases.
        """
        model = NMF(n_components=self.K, init="random", max_iter=200)
        P = model.fit_transform(X)
        Q = model.components_
        self.pred = np.dot(P, Q)
        self.invP = None
        return self



from holisticai.bias.mitigation.inprocessing.commons import Logging


class RBLogger:
    def __init__(self, total_iterations, verbose):
        self.verbose = verbose
        log_params = [
            ("iteration", int),
            ("primal_residual:", float),
            ("dual_residual:", float),
        ]
        self.logger = Logging(
            log_params=log_params,
            total_iterations=total_iterations,
            logger_format="iteration",
        )

    def update(self, iteration, primal_residual, dual_residual):
        if self.verbose:
            self.logger.update(iteration, primal_residual, dual_residual)



import numpy as np


class ActiveLearningItemsSelection:
    def __init__(self, top_n=10, theta=3):
        self.top_n = top_n
        self.theta = theta

    def __call__(self, time_mask, pred_data, return_only_time_mask=False):
        pred_data = 5 - np.abs(self.theta - pred_data)
        candidate_index = ~time_mask  ## convert all candidate index to True
        candidate_rating = (
            pred_data * candidate_index
        )  ## get predicte rating on those candidate index
        sorted_ind = np.argsort(-candidate_rating, axis=1)[
            :, :
        ]  ## get the index of each row, where top n pred are selected
        items_selected_matrix = []
        for row, selected_index in enumerate(sorted_ind):
            candidate_index[row, selected_index] = False
            items_selected_matrix.append(selected_index)
        items_selected_matrix = np.array(items_selected_matrix)
        time_mask = ~candidate_index  ## get the new index
        if return_only_time_mask:
            return time_mask
        else:
            return items_selected_matrix, time_mask


class ConventionalItemsSelection:
    def __init__(self, top_n=10):
        self.top_n = top_n

    def __call__(self, time_mask, pred_data, return_only_time_mask=False):
        candidate_index = ~time_mask  ## convert all candidate index to True
        candidate_rating = (
            pred_data * candidate_index
        )  ## get predicte rating on those candidate index
        sorted_ind = np.argsort(-candidate_rating, axis=1)[
            :, : self.top_n
        ]  ## get the index of each row, where top n pred are selected
        items_selected_matrix = []
        for row, selected_index in enumerate(sorted_ind):
            candidate_index[row, selected_index] = False
            items_selected_matrix.append(selected_index)
        items_selected_matrix = np.array(items_selected_matrix)
        time_mask = ~candidate_index  ## get the new index
        if return_only_time_mask:
            return time_mask
        else:
            return items_selected_matrix, time_mask


class RandomItemSelection:
    def __init__(self, top_n=10):
        self.top_n = top_n

    def __call__(self, time_mask, pred_data, return_only_time_mask=False):
        candidate_index = ~time_mask  ## convert all candidate index to True
        candidate_rating = (
            pred_data * candidate_index
        )  ## get predicte rating on those candidate index
        sorted_ind = np.argsort(-candidate_rating, axis=1)[
            :, :
        ]  ## get the index of each row, where top n pred are selected
        items_selected_matrix = []
        for row, selected_index in enumerate(sorted_ind):
            item_selected = np.random.choice(selected_index, self.top_n, replace=False)
            candidate_index[row, item_selected] = False
            items_selected_matrix.append(item_selected)
        items_selected_matrix = np.array(items_selected_matrix)
        time_mask = ~candidate_index  ## get the new index
        if return_only_time_mask:
            return time_mask
        else:
            return items_selected_matrix, time_mask



import numpy as np
from scipy.optimize import linprog
from scipy.sparse import lil_matrix
from tqdm import trange

from holisticai.utils.transformers.bias import SensitiveGroups


class Algorithm:
    def __init__(self, metric, solver, verbose):
        self.metric = metric
        self.solver = solver
        self.verbose = verbose
        self.sens_group = SensitiveGroups()

    def init_parameters(self, y_pred: np.ndarray, p_attr: np.ndarray):

        k = len(np.unique(y_pred))

        n_classes = np.max(y_pred) + 1
        y_oh = np.zeros((y_pred.size, n_classes))
        y_oh[np.arange(y_pred.size), y_pred] = 1
        p_attr = np.array(p_attr).reshape(-1)
        Nx = np.sum(p_attr)
        q = Nx // k
        r = Nx - q * k
        lower_bound = q
        if r >= 1:
            upper_bound = q + 1
        else:
            upper_bound = q

        P = p_attr.astype("int32")
        n = len(P)
        return {
            "lower_bound": lower_bound,
            "upper_bound": upper_bound,
            "len_data": n,
            "y_oh": y_oh,
            "P": P,
            "k": k,
        }

    def penalty_weights(self, X=None, centroids: np.ndarray = None):
        if self.metric == "constant":
            return -1
        elif self.metric == "L1":
            norm_p = 1
            d = np.linalg.norm(
                centroids[:, None, :] - X[None, ...], ord=norm_p, axis=-1
            )
            w = (-d).reshape(-1)
            return w
        elif self.metric == "L2":
            norm_p = 2
            d = np.linalg.norm(
                centroids[:, None, :] - X[None, ...], ord=norm_p, axis=-1
            )
            w = (-d).reshape(-1)
            return w
        else:
            raise Exception(f"Unknown Measure : {self.metric}")

    def compute_cost_function(self, centroids, X, z_pred, z_mod):
        if self.metric == "constant":
            return 1 - np.sum(z_mod * z_pred, axis=0)
        elif self.metric == "L1":
            norm_p = 1
            d = np.linalg.norm(
                centroids[:, None, :] - X[None, ...], ord=norm_p, axis=-1
            )
            w = np.mean(d, axis=0) - np.sum(d * z_mod, axis=0)
            return np.sum(w * (1 - np.sum(z_mod * z_pred, axis=0)))
        elif self.metric == "L2":
            norm_p = 2
            d = np.linalg.norm(
                centroids[:, None, :] - X[None, ...], ord=norm_p, axis=-1
            )
            w = np.mean(d, axis=0) - np.sum(d * z_mod, axis=0)
            return np.sum(w * (1 - np.sum(z_mod * z_pred, axis=0)))
        else:
            raise Exception(f"Unknown Measure : {self.metric}")

    def transform(
        self,
        X: np.ndarray,
        y_pred: np.ndarray,
        group: np.ndarray,
        centroids: np.ndarray = None,
    ):

        params = self.init_parameters(y_pred, group)
        n = params["len_data"]
        k = params["k"]
        P = params["P"]
        upper_bound = params["upper_bound"]
        lower_bound = params["lower_bound"]

        weights = self.penalty_weights(X=X, centroids=centroids)
        zi = (-params["y_oh"].T).reshape(-1)
        A = -np.concatenate([weights * zi, np.zeros(2 * k)])

        ncons = 2 * k + n
        nvars = k * n + 2 * k
        C = lil_matrix((ncons, nvars))
        p = trange(k) if self.verbose > 0 else range(k)
        for i in p:
            Pindex = np.where(P == 1)[0]
            C[i, Pindex + i * n] = 1
            C[i, k * n + i] = 1

        p = trange(k) if self.verbose > 0 else range(k)
        for i in p:
            C[i + k, Pindex + i * n] = -1
            C[i + k, k * n + k + i] = 1

        p = trange(n) if self.verbose > 0 else range(n)
        for i in p:
            C[2 * k + i, i + np.arange(k) * n] = 1

        obj = A
        lhs_eq = C
        rhs_eq = np.concatenate(
            [np.ones(k) * upper_bound, -np.ones(k) * lower_bound, np.ones(n)]
        ).reshape([-1, 1])

        bnd = [(0, 1)] * nvars
        opt = linprog(c=obj, A_eq=lhs_eq, b_eq=rhs_eq, bounds=bnd, method=self.solver)
        new_y_pred = opt.x[: k * n].reshape([k, n]).argmax(axis=0)
        if self.verbose > 0:
            self.cost = self.compute_cost_function(centroids, X, y_pred, new_y_pred)
            print(self.cost)
        return new_y_pred



import numpy as np

from .algorithm_utils import Utils


class OPT_Modification:
    def __init__(self, k, verbose=0):
        self.k = k
        self.verbose = verbose

    def fit(self, y_pred, p_attr):
        y_pred = self.run(y_pred, p_attr)

    def run(self, y_pred, p_attr):
        Nx, q, r, lower_bound, upper_bound, Cj = self.init_parameters(y_pred, p_attr)
        self.utils = Utils(
            upper_bound=upper_bound, lower_bound=lower_bound, Nx=Nx, k=self.k
        )

        if r == 0:
            print("Case 1")
            gt, gp = [], []
            for j in range(self.k):
                if Cj[j]["beta"] > Nx // self.k:
                    gt.append(j)
                if Cj[j]["beta"] <= Nx // self.k:
                    gp.append(j)
            T, Cj = self.utils.select_items(gt, Cj, mode="exact")
            T, Cj = self.utils.choose_items(gp, Cj, T, mode="exact")
        else:
            g1, g2, g3, g4 = [], [], [], []
            for j in range(self.k):
                if Cj[j]["beta"] > upper_bound:
                    g1.append(j)
                if Cj[j]["beta"] == upper_bound:
                    g2.append(j)
                if Cj[j]["beta"] == lower_bound:
                    g3.append(j)
                if Cj[j]["beta"] < lower_bound:
                    g4.append(j)

            t = len(g1)
            p = len(g1 + g2)
            m = len(g1 + g2 + g3)

            gs = g1 + g2 + g3 + g4
            print("Case 2")
            if p < r:
                print("Case 2.1")
                T, Cj = self.utils.select_items(g1, Cj, mode="upper")
                rp = r - p
                g = g3 + g4
                T, Cj = self.utils.choose_items(g[:rp], Cj, T, mode="upper")
                T, Cj = self.utils.choose_items(g[rp:], Cj, T, mode="lower")
            elif p == r:
                print("Case 2.2")
                T, Cj = self.utils.select_items(g1, Cj, mode="upper")
                T, Cj = self.utils.choose_items(g4, Cj, T, mode="lower")
            elif p > r:
                print("Case 2.3")
                if t > r:
                    print("Case 2.3.1")
                    T, Cj = self.utils.select_items(g1, Cj, mode="upper")
                    T, Cj = self.utils.select_items(gs[r:p], Cj, mode="lower")
                    T, Cj = self.utils.choose_items(g4, Cj, T, mode="lower")

                elif t == r:
                    print("Case 2.3.2")
                    T, Cj = self.utils.select_items(g1, Cj, mode="upper")
                    T, Cj = self.utils.select_items(g2, Cj, mode="lower")
                    T, Cj = self.utils.choose_items(g4, Cj, T, mode="lower")

                elif t < r:
                    print("Case 2.3.3")
                    T, Cj = self.utils.select_items(g1, Cj, mode="upper")
                    start = r - t
                    end = p
                    T, Cj = self.utils.select_items(gs[start:end], Cj, mode="lower")
                    T, Cj = self.utils.choose_items(g4, Cj, T, mode="lower")

        for js in range(self.k):
            jr = Cj[js]["jr"]
            index = Cj[js]["index"]
            y_pred[index] = jr

        if self.verbose > 0:
            for js in range(self.k):
                print(Cj[js]["beta"])

        return y_pred, T

    def init_parameters(self, y_pred, p_attr):
        Nx = np.sum(p_attr == 1)
        q = Nx // self.k
        r = Nx - q * self.k

        if r >= 1:
            lower_bound = q
            upper_bound = q + 1
        else:
            lower_bound = upper_bound = q

        cluster_info = []
        for j in range(self.k):
            c_index = np.where((y_pred == j) & (p_attr == 1))[0]
            beta_j = len(c_index)
            cluster_info.append((j, beta_j, c_index))

        Cj = {}
        for js, (jr, beta, index) in enumerate(
            sorted(cluster_info, key=lambda x: x[1], reverse=True)
        ):
            Cj[js] = {"jr": jr, "beta": beta, "index": list(index)}

        return Nx, q, r, lower_bound, upper_bound, Cj



# imports
from .calibrated_eq_odds_postprocessing import CalibratedEqualizedOdds
from .debiasing_exposure.transformer import DebiasingExposure
from .eq_odds_postprocessing import EqualizedOdds
from .fair_topk.transformer import FairTopK
from .lp_debiaser.binary_balancer.transformer import LPDebiaserBinary
from .lp_debiaser.multiclass_balancer.transformer import LPDebiaserMulticlass
from .mcmf_clustering.transformer import MCMF
from .ml_debiaser.transformer import MLDebiaser
from .plugin_estimator_and_recalibration.transformer import (
    PluginEstimationAndCalibration,
)
from .reject_option_classification import RejectOptionClassification
from .wasserstein_barycenters.transformer import WassersteinBarycenter

__all__ = [
    "CalibratedEqualizedOdds",
    "EqualizedOdds",
    "RejectOptionClassification",
    "WassersteinBarycenter",
    "PluginEstimationAndCalibration",
    "MLDebiaser",
    "LPDebiaserBinary",
    "LPDebiaserMulticlass",
    "DebiasingExposure",
    "FairTopK",
    "MCMF",
]

import importlib

networkx_spec = importlib.util.find_spec("networkx")
if networkx_spec is not None:
    from .disparate_impact_remover_rs import DisparateImpactRemoverRS

__all__ += ["DisparateImpactRemoverRS"]



"""
The :mod:`holisticai.bias.plots` module includes bias plotters.
"""
# Classification Plots
from ._bias_classification_plots import abroca_plot

# Exploratory plots
from ._bias_exploratory_plots import distribution_plot, group_pie_plot, histogram_plot

# Multiclass Plots
from ._bias_multiclass_plots import (
    accuracy_bar_plot,
    disparate_impact_plot,
    frequency_matrix_plot,
    frequency_plot,
    statistical_parity_plot,
)

# Recommender Plots
from ._bias_recommender_plots import (
    exposure_diff_plot,
    exposure_ratio_plot,
    long_tail_plot,
)

# Regression Plots
from ._bias_regression_plots import (
    disparate_impact_curve,
    mae_bar_plot,
    rmse_bar_plot,
    statistical_parity_curve,
    success_rate_curve,
    success_rate_curves,
)

# All bias plotters
__all__ = [
    "abroca_plot",
    "success_rate_curve",
    "disparate_impact_curve",
    "statistical_parity_curve",
    "success_rate_curves",
    "rmse_bar_plot",
    "mae_bar_plot",
    "exposure_diff_plot",
    "exposure_ratio_plot",
    "frequency_plot",
    "statistical_parity_plot",
    "disparate_impact_plot",
    "frequency_matrix_plot",
    "accuracy_bar_plot",
    "long_tail_plot",
    "group_pie_plot",
    "distribution_plot",
    "histogram_plot",
]



import numpy as np
import pandas as pd


def cp_mat(y_true, y_pred, n_classes):
    """Returns the matrix of conditional probabilities y_pred | y_true"""
    tab = pd.crosstab(y_true, y_pred)
    cols = list(set(range(n_classes)) - set(tab.columns))
    if not cols == []:
        tab[cols] = 0
    tab = tab.values
    probs = tab.transpose() / tab.sum(axis=1)
    return probs.transpose()


def p_vec(y, flatten=True):
    """Returns the matrix of probabilities for the levels y"""
    tab = pd.crosstab(y, "count").values
    out = tab / tab.sum()
    if flatten:
        out = out.flatten()
    return out


def pars_to_cpmat(opt, n_groups=3, n_classes=3):
    """Reshapes the LP parameters as an n_group * n_class * n_class array"""
    shaped = np.reshape(opt.x, (n_groups, n_classes, n_classes))
    flipped = np.array([m.T for m in shaped])
    return flipped



import numpy as np


class PRBinaryCrossEntropy:
    def __init__(self, eta, C):
        self.eta = eta
        self.C = C
        self.freg = BinaryFairnessRegularizer()
        self.xent = BinaryCrossEntropy()

    def __call__(self, y, sigma, groups, coef):
        l = self.xent(y, sigma)
        f = self.freg(sigma, groups)
        # l2 regularizer
        reg = np.sum(coef * coef)
        loss = -l + self.eta * f + 0.5 * self.C * reg
        return loss

    def gradient(self, X, y, sigma, groups, coef):
        l = self.xent.gradient(X, y, sigma, groups)
        f = self.freg.gradient(X, sigma, groups)
        # l2 regularizer
        reg = coef
        # sum
        loss = np.reshape(-l + self.eta * f + self.C * reg, [-1])
        return loss


class BinaryFairnessRegularizer:
    def __call__(self, sigma, groups):
        # fairness-aware regularizer
        # \sum_{x,s in D} \
        #    sigma(x,x)       [log(rho(s))     - log(pi)    ] + \
        #    (1 - sigma(x,s)) [log(1 - rho(s)) - log(1 - pi)]
        rho_s, pi = self._parameters(sigma, groups)
        return (
            sigma * (np.log(rho_s) - np.log(pi))
            + (1.0 - sigma) * (np.log(1.0 - rho_s) - np.log(1.0 - pi))
        ).sum()

    def _parameters(self, sigma, groups):
        # rho(s) = Pr[y=0|s] = \sum_{(xi,si)in D st si=s} sigma(xi,si) / #D[s]
        # d_rho(s) = \sum_{(xi,si)in D st si=s} d_sigma(xi,si) / #D[s]
        sigma_group = op_by_group(sigma, groups, reduce_op=np.mean)

        rho_s = sigma_group[groups]

        # pi = Pr[y=0] = \sum_{(xi,si)in D} sigma(xi,si) / #D
        pi = sigma.mean()

        return rho_s, pi

    def _derivate_parameters(self, X, sigma, groups):
        d_sigma = (sigma * (1.0 - sigma))[:, np.newaxis] * X

        # rho(s) = Pr[y=0|s] = \sum_{(xi,si)in D st si=s} sigma(xi,si) / #D[s]
        # d_rho(s) = \sum_{(xi,si)in D st si=s} d_sigma(xi,si) / #D[s]
        d_sigma_group = op_by_group(d_sigma, groups, reduce_op=np.mean, axis=0)

        d_rho_s = d_sigma_group[groups]

        # pi = Pr[y=0] = \sum_{(xi,si)in D} sigma(xi,si) / #D
        d_pi = d_sigma.mean(axis=0)

        return d_sigma, d_rho_s, d_pi

    def gradient(self, X, sigma, groups):
        # fairness-aware regularizer
        # differentialy by w(s)
        # \sum_{x,s in {D st s=si} \
        #     [(log(rho(si)) - log(pi)) - (log(1 - rho(si)) - log(1 - pi))] \
        #     * d_sigma
        # + \sum_{x,s in {D st s=si} \
        #     [ {sigma(xi, si) - rho(si)} / {rho(si) (1 - rho(si))} ] \
        #     * d_rho
        # - \sum_{x,s in {D st s=si} \
        #     [ {sigma(xi, si) - pi} / {pi (1 - pi)} ] \
        #     * d_pi
        rho_s, pi = self._parameters(sigma, groups)
        d_sigma, d_rho_s, d_pi = self._derivate_parameters(X, sigma, groups)

        f1 = (np.log(rho_s) - np.log(pi)) - (np.log(1.0 - rho_s) - np.log(1.0 - pi))
        f2 = (sigma - rho_s) / (rho_s * (1.0 - rho_s))
        f3 = (sigma - pi) / (pi * (1.0 - pi))

        f4 = (f1[:, np.newaxis] * d_sigma + f2[:, np.newaxis] * d_rho_s) - np.outer(
            f3, d_pi
        )
        f = op_by_group(f4, groups, reduce_op=np.sum, axis=0)
        return f


class BinaryCrossEntropy:
    def __call__(self, y, sigma):
        # likelihood
        # \sum_{x,s,y in D} y log(sigma) + (1 - y) log(1 - sigma)
        return (y * np.log(sigma) + (1.0 - y) * np.log(1.0 - sigma)).sum()

    def gradient(self, X, y, sigma, groups):
        # likelihood
        # l(si) = \sum_{x,y in D st s=si} (y - sigma(x, si)) x
        loss = (y - sigma)[:, np.newaxis] * X
        l = op_by_group(loss, groups, reduce_op=np.sum, axis=0)
        return l


def op_by_group(matrix, groups, reduce_op, axis=None):
    nb_group_values = len(np.unique(groups))
    result = [reduce_op(matrix[groups == i], axis=axis) for i in range(nb_group_values)]
    return np.stack(result)



import numpy as np
from sklearn.pipeline import Pipeline as SKLPipeline
from sklearn.utils.metaestimators import available_if

from holisticai.pipeline.handlers._estimator import EstimatorHandler
from holisticai.pipeline.handlers._pipeline_params import PipelineParametersHandler
from holisticai.pipeline.handlers._utransformers import UTransformersHandler

SUPPORTED_FUNCTIONS = [
    "fit",
    "predict",
    "predict_proba",
    "predict_score",
    "predictions",
]
POST_PREDICTION = {
    "predict": "y_pred",
    "predict_score": "y_score",
    "predict_proba": "y_proba",
}
POST_SCORE = "score"


class PipelineHelper:
    def __getattribute__(self, name):
        """
        This function return the attribute using the following rules:
        - for all functions listed in SUPPORTED_FUNCTIONS return a wrapped function that update the input arguments with the cache before call fit function.
        - for any other function we check if the pipeline has the attribute else try the attribute from the main class.
        """
        if name in SUPPORTED_FUNCTIONS:
            return object.__getattribute__(self, "handle_pipeline_methods")(name)
        else:
            return object.__getattribute__(self, name)

    def handle_pipeline_methods(self, fn_name):
        def function(X, y=None, **kargs):
            params = self.preprocessing_parameters(kargs, y)

            if self.post_estimator_transformers:

                if fn_name in POST_PREDICTION:
                    # For predictions we must check supported methods
                    fn = object.__getattribute__(self, "predictions")
                    output = fn(X, **params)[POST_PREDICTION[fn_name]]

                elif fn_name == POST_SCORE:
                    # For score we must check supported methods
                    output = fn = object.__getattribute__(self, "score")

                else:
                    # For fitting model the method mus be called from the main class
                    output = object.__getattribute__(self, fn_name)(X, **params)

            else:
                # Other case the primitive method could be invoked
                output = getattr(SKLPipeline, fn_name)(self, X, **params)

            return output

        return function

    @property
    def post_estimator_transformers(self):
        """
        return list o post-estimator transformers
        """
        return self.utransformers_hdl.pos_processing_steps

    def preprocessing_steps(self, steps):
        """
        Preprocessing steps before calling sklearn.pipeline.Pipeline contructor

        Description
        ----------
        This function collects all bias mitigation steps, remove steps after estimator and wrap the estimator

        Parameters
        ----------
        steps : list
            pipeline steps

        Returns
        -------
        list
            sklearn pipeline steps

        """
        self.params_hdl = PipelineParametersHandler()
        self.estimator_hdl = EstimatorHandler(self.params_hdl)
        self.utransformers_hdl = UTransformersHandler(
            steps, self.params_hdl, self.estimator_hdl
        )

        steps = self.utransformers_hdl.drop_post_processing_steps(steps)
        steps = self.estimator_hdl.wrap_and_link_estimator_step(steps)

        return steps

    def preprocessing_parameters(self, params, y=None):
        """
        Preprocessing parameters.

        Description
        ----------
        The function filter, drop and feed parameters for unconventional transformers.

        Parameters
        ----------
        params : dict
            dictionary of parameters

        Returns
        -------
        dict
            estimator parameters

        """
        params = self.params_hdl.bias_mitigator.feed(params, return_dropped=True)
        self.params_hdl.estimator.feed(params)

        if y is not None:
            params.update({"y": y})
            self.params_hdl.estimator["y"] = y

        if "sample_weight" in params:
            self.params_hdl.estimator["sample_weight"] = params["sample_weight"]

        return params

    def fit_post_estimator_transformers(self, Xt, y):
        """
        fit post-estimator transformers.

        Description
        ----------
        Complete the pipeline fit procedure for transformers defined after the estimator in the pipeline.
        Prediction is necessary to avoid conflicts with preprocessing approaches during fitting.

        Parameters
        ----------
        Xt : numpy array
            Transformed input matrix

        y : numpy array
            Target vector

        Returns
        -------
        None
        """
        output_kargs = self.estimator_hdl.run_predictions(Xt)
        self.utransformers_hdl.fit_postprocessing(Xt, y, **output_kargs)

    def _transform_post_estimator_transformers(self, Xt, **params):
        """
        call transform for post-estimator transformers.

        Description
        ----------
        Complete the pipeline transform procedure for transformers defined after the estimator in the pipeline.

        Parameters
        ----------
        Xt : numpy array
            Transformed input matrix

        Returns
        -------
        dict
            Post-processed predictions vectors
        """

        output_kargs = self.estimator_hdl.run_predictions(Xt)
        transform_kargs = {k[4:]: v for k, v in params.items() if k.startswith("bm__")}
        transform_kargs["X"] = Xt
        output_kargs.update(transform_kargs)
        return self.utransformers_hdl.transform_postprocessing(**output_kargs)

    def _transform_without_final(self, X):
        """Transform the data. Not apply `transform` with the final estimator.

        Call `transform` of each transformer in the pipeline. The transformed
        data are finally passed to the final estimator that calls
        `transform` method. Only valid if the final estimator
        implements `transform`.

        This also works where final estimator is `None` in which case all prior
        transformations are applied.

        Parameters
        ----------
        X : matrix-like
            Data to transform. Must fulfill input requirements of first step
            of the pipeline.

        Returns
        -------
        Xt : ndarray of shape (n_samples, n_transformed_features)
            Transformed data.
        """
        Xt = X
        for _, _, transform in self._iter(with_final=False):
            Xt = transform.transform(Xt)
        return Xt



from typing import Optional

import numpy as np
import pandas as pd

from .._conventions import (
    _ALL,
    _EVENT,
    _GROUP_ID,
    _LABEL,
    _LOWER_BOUND_DIFF,
    _PRED,
    _SIGNED,
    _UPPER_BOUND_DIFF,
)
from .._moments_utils import BaseMoment, format_data
from ._objectives import ErrorRate


class ClassificationConstraint(BaseMoment):
    """Extend Base Moment for problem that can be expressed as weighted classification error."""

    PROBLEM_TYPE = "classification"

    def __init__(self, ratio_bound: Optional[float] = 1.0):
        """
        Initialize with the ratio value.

        Parameters
        ----------
        ratio_bound : float
            The constraints' ratio bound for constraints that are expressed as
            ratios. The specified value needs to be in (0,1].
        """

        self.eps = 0.001
        self.ratio = ratio_bound

    def load_data(
        self,
        X,
        y,
        sensitive_features: pd.Series,
        event: pd.Series,
        utilities: np.ndarray = None,
    ):
        """
        Description
        -----------
        Load the specified data into this object.

        The `utilities` is a 2-d array which corresponds to g(X,A,Y,h(X)) [agarwal2018reductions]
        The `utilities` defaults to h(X), i.e. [0, 1] for each X_i.

        The first column is G^0 and the second is G^1.
        Assumes binary classification with labels 0/1.

        .. math::
            utilities = [g(X,A,Y,h(X)=0), g(X,A,Y,h(X)=1)]
        """
        if utilities is None:
            utilities = np.vstack(
                [
                    np.zeros(y.shape, dtype=np.float64),
                    np.ones(y.shape, dtype=np.float64),
                ]
            ).T
        self.utilities = utilities
        super().load_data(X, y, sensitive_features)
        self._build_event_variables(event)

    def _build_event_variables(self, event):
        """
        This method:
        - adds a column `event` to the `tags` field.
        - fill in the information about the basis
        """
        self.tags[_EVENT] = event

        # Events
        self.event_ids = np.sort(self.tags[_EVENT].dropna().unique())
        self.event_prob = self.tags[_EVENT].dropna().value_counts() / len(self.tags)

        # Groups and Events
        self.group_values = np.sort(self.tags[_GROUP_ID].unique())
        self.group_event_prob = (
            self.tags.dropna(subset=[_EVENT]).groupby([_EVENT, _GROUP_ID]).count()
            / len(self.tags)
        ).iloc[:, 0]

        self.index = self._get_index_format()
        self.default_objective_lambda_vec = None
        self._get_basis()

    def signed_weights(self, lambda_vec):
        """
        Compute the signed weights.

        Description
        -----------

        Uses the equations for :math:`C_i^0` and :math:`C_i^1` as defined
        in Section 3.2 of :footcite:t:`agarwal2018reductions`
        in the 'best response of the Q-player' subsection to compute the
        signed weights to be applied to the data by the next call to the underlying
        estimator.

        Parameters
        ----------
        lambda_vec : :class:`pandas:pandas.Series`
            The vector of Lagrange multipliers indexed by `index`
        """

        lambda_event = (lambda_vec["+"] - self.ratio * lambda_vec["-"]).sum(
            level=_EVENT
        ) / self.event_prob

        lambda_group_event = (
            self.ratio * lambda_vec["+"] - lambda_vec["-"]
        ) / self.group_event_prob

        adjust = lambda_event - lambda_group_event

        def get_signed_weight(row):
            if pd.isna(row[_EVENT]):
                return 0
            else:
                return adjust[row[_EVENT], row[_GROUP_ID]]

        signed_weights = self.tags.apply(get_signed_weight, axis=1)
        utility_diff = self.utilities[:, 1] - self.utilities[:, 0]
        signed_weights = utility_diff.T * signed_weights
        return signed_weights

    def default_objective(self):
        return ErrorRate()

    def gamma(self, predictor):
        """Calculate the degree to which constraints are currently violated by the predictor."""

        utility_diff = self.utilities[:, 1] - self.utilities[:, 0]
        predictions = np.squeeze(predictor(self.X))
        pred = utility_diff.T * predictions + self.utilities[:, 0]

        return self._gamma_signed(pred)

    def _gamma_signed(self, pred):
        self.tags[_PRED] = pred
        expect_event = self.tags.groupby(_EVENT).mean()
        expect_group_event = self.tags.groupby([_EVENT, _GROUP_ID]).mean()
        expect_group_event[_UPPER_BOUND_DIFF] = (
            self.ratio * expect_group_event[_PRED] - expect_event[_PRED]
        )
        expect_group_event[_LOWER_BOUND_DIFF] = (
            -expect_group_event[_PRED] + self.ratio * expect_event[_PRED]
        )
        gamma_signed = pd.concat(
            [
                expect_group_event[_UPPER_BOUND_DIFF],
                expect_group_event[_LOWER_BOUND_DIFF],
            ],
            keys=["+", "-"],
            names=["signed", _EVENT, _GROUP_ID],
        )
        return gamma_signed

    def _get_basis(self):
        pos_basis = pd.DataFrame()
        neg_basis = pd.DataFrame()
        neg_basis_present = pd.Series(dtype="float64")
        zero_vec = pd.Series(0.0, self.index)
        i = 0
        for event_val in self.event_ids:
            for group in self.group_values[:-1]:
                pos_basis[i] = zero_vec
                neg_basis[i] = zero_vec
                pos_basis[i]["+", event_val, group] = 1
                neg_basis[i]["-", event_val, group] = 1
                neg_basis_present.at[i] = True
                i += 1
        self.neg_basis_present = neg_basis_present
        self.basis = {"+": pos_basis, "-": neg_basis}

    def _get_index_format(self):
        index = (
            pd.DataFrame(
                [
                    {_SIGNED: signed, _EVENT: e, _GROUP_ID: g}
                    for e in self.event_ids
                    for g in self.group_values
                    for signed in ["+", "-"]
                ]
            )
            .set_index([_SIGNED, _EVENT, _GROUP_ID])
            .index
        )
        return index


class DemographicParity(ClassificationConstraint):
    """
    Extend ClassificationConstraint for demographic parity constraint.

    A classifier :math:`h(X)` satisfies demographic parity if

      P[h(X) = 1 | A = a] = P[h(X) = 1] \; \forall a

    - base_event : defines a single event, `ALL`.
    - event_prob : will record only the probability 1.
    """

    short_name = "DemographicParity"

    def load_data(self, X, y, sensitive_features):
        """Load the specified data into the object."""
        params = format_data(y=y)
        y = params["y"]
        base_event = pd.Series(data=_ALL, index=y.index)
        super().load_data(X, y, sensitive_features, base_event)


class EqualizedOdds(ClassificationConstraint):
    """
    Extend ClassificationConstraint for Equalized Odds constraint.

    A classifier :math:`h(X)` satisfies equalized odds if
    .. math::
       P[h(X) = 1 | A = a, Y = y] = P[h(X) = 1 | Y = y] \; \forall a, y

    - base_event: defines the event corresponding to unique value in the `Y` array.
    - event_prob: will record the fraction of the samples corresponding to each unique value in
    the `Y` array.
    """

    short_name = "EqualizedOdds"

    def load_data(self, X, y, sensitive_features):
        """Load the specified data into the object."""
        params = format_data(y=y)
        y = params["y"]
        base_event = y.apply(lambda v: _LABEL + "=" + str(v))
        super().load_data(X, y, sensitive_features, base_event)


class TruePositiveRateParity(ClassificationConstraint):
    """
    Extend ClassificationConstraint for true positive rate parity (equal opportunity) constraint.

    A classifier :math:`h(X)` satisfies true positive rate parity if
    .. math::
       P[h(X) = 1 | A = a, Y = 1] = P[h(X) = 1 | Y = 1] \; \forall a

    - base_event: defines the event corresponding to Y = 1 .
    - event_prob: will record the fraction of the samples corresponding to `Y = 1` in the `Y` array.
    """

    short_name = "TruePositiveRateParity"

    def load_data(self, X, y, sensitive_features):
        """Load the specified data into the object."""
        params = format_data(y=y)
        y = params["y"]
        base_event = y.apply(lambda l: f"{_LABEL}={l}").where(y == 1)
        super().load_data(X, y, sensitive_features, base_event)


class FalsePositiveRateParity(ClassificationConstraint):
    """
    Extend ClassificationConstraint for false positive rate parity constraint.

    A classifier :math:`h(X)` satisfies false positive rate parity if
    .. math::
       P[h(X) = 1 | A = a, Y = 0] = P[h(X) = 1 | Y = 0] \; \forall a

    - base_event: defines the event corresponding to Y = 0 .
    - event_prob: will record the fraction of the samples corresponding to `Y = 0` in the `Y` array.
    """

    short_name = "FalsePositiveRateParity"

    def load_data(self, X, y, sensitive_features):
        """Load the specified data into the object."""
        params = format_data(y=y)
        y = params["y"]
        base_event = y.apply(lambda v: _LABEL + "=" + str(v)).where(y == 0)
        super().load_data(X, y, sensitive_features, base_event)


class ErrorRateParity(ClassificationConstraint):
    """
    Extend ClassificationConstraint for error rate parity constraint.

    A classifier :math:`h(X)` satisfies error rate parity if
    .. math::
       P[h(X) \ne Y | A = a] = P[h(X) \ne Y] \; \forall a

    - base_event : defines a single event, `ALL`.
    - event_prob : will record only the probability 1.
    - In this scenario, g = abs(h(x)-y), rather than g = h(x)
    """

    short_name = "ErrorRateParity"

    def load_data(self, X, y, sensitive_features):
        """Load the specified data into the object."""
        params = format_data(y=y)
        y = params["y"]
        utilities = np.vstack([y, 1 - y]).T
        base_event = pd.Series(data=_ALL, index=y.index)
        super().load_data(X, y, sensitive_features, base_event, utilities=utilities)



import numpy as np


def normalize(S_in):
    maxcol = np.max(S_in, axis=1, keepdims=True)
    S_in = S_in - maxcol
    S_out = np.exp(S_in)
    S_out_sum = np.sum(S_out, axis=1, keepdims=True)
    S_out = S_out / S_out_sum

    return S_out


def normalize_2(S_in):
    S_in_sum = np.sum(S_in, axis=1, keepdims=True)
    S_in = S_in / S_in_sum
    return S_in


def bound_energy(S, S_in, a_term, b_term):
    E = np.nansum(
        (
            S * np.log(np.maximum(S, 1e-15))
            - S * np.log(np.maximum(S_in, 1e-15))
            + a_term * S
            + b_term * S
        )
    )
    return E


def get_S_discrete(l, N, K):
    x = range(N)
    temp = np.zeros((N, K), dtype=float)
    temp[(x, l)] = 1
    return temp


def compute_b(S, groups_ids, group_prob, F_a):
    R = (group_prob.values[:, None] / np.sum(S, axis=0, keepdims=True)).T

    V = groups_ids.values[:, None, :]

    F_b = np.maximum(np.sum(V * S[:, :, None], axis=0, keepdims=True), 1e-15)

    F_a_b = F_a[:, None, :] / F_b

    return np.sum(R[None, :, :] - F_a_b, axis=-1)


class BoundUpdate:
    def __init__(self, bound_lambda, L, bound_iteration):
        self.bound_lambda = bound_lambda
        self.L = L
        self.bound_iteration = bound_iteration

    def transform(self, a_p, group_prob, groups_ids):
        oldE = float("inf")
        S = np.exp((-a_p))
        S = normalize_2(S)
        a_term = -a_p.copy()

        F_a = group_prob.values[None, :] * groups_ids.values

        for i in range(self.bound_iteration):

            b = compute_b(S, groups_ids, group_prob, F_a)
            b_term = self.bound_lambda * b
            terms = (a_term - b_term) / self.L
            S_in_2 = normalize(terms)
            S_in = S.copy()
            S = S * S_in_2
            S = normalize_2(S)

            E = bound_energy(S, S_in, a_p, b_term)

            report_E = E

            if i > 1 and (abs(E - oldE) <= 1e-5 * abs(oldE)):
                break

            else:
                oldE = E
                report_E = E

        l = np.argmax(S, axis=1)

        return l, S, report_E



import os
import sys

sys.path.append(os.getcwd())
import numpy as np

from holisticai.utils.transformers.bias import SensitiveGroups

from .algorithm_utils import f_lambda


class PluginEstimationAndCalibrationAlgorithm:
    def __init__(self, L=25, beta=0.1):
        self.M = 1
        self.L = np.floor(L / 2).astype(np.int32)
        self.beta = beta
        self.sens_groups = SensitiveGroups()
        self.eps = np.finfo(float).eps

    def fit(self, y_pred: np.ndarray, sensitive_features: np.ndarray):
        SL = self.sens_groups.fit_transform(sensitive_features, convert_numeric=True)
        YL = y_pred * 2 - 1
        self.p = [np.mean(SL == 0), np.mean(SL == 1)]
        self.lambda_ = f_lambda(YL, SL, self.M, self.L, self.beta)

    def transform(self, y_pred: np.ndarray, sensitive_features: np.ndarray):
        ST = self.sens_groups.transform(sensitive_features, convert_numeric=True)
        YT = y_pred * 2 - 1
        i = np.arange(-self.L, self.L + 1, 1)
        YTd = np.zeros(len(YT))
        for k in range(len(YT)):
            v = (
                self.p[ST[k]] * (YT[k] - i * self.M / self.L) ** 2
                + (1 - 2 * ST[k]) * self.lambda_
            )
            j = np.argmin(v)
            YTd[k] = i[j] * self.M / self.L
        YTd = (YTd + 1) / 2
        return YTd



import logging

import numpy as np
import pandas as pd

from .algorithm_utils._fair_clustering import FairClustering
from .algorithm_utils._utils import (
    get_fair_accuracy,
    get_fair_accuracy_proportional,
    normalizefea,
)


class FairClusteringAlgorithm:
    def __init__(
        self, K, L, lmbda, method="kmeans", normalize_input=True, verbose=True
    ):
        self.lmbda = lmbda
        self.K = K
        self.L = L
        self.normalize_input = normalize_input
        self.fair_clustering = FairClustering(
            K, L, lmbda, method=method, verbose=verbose
        )

    def fit_transform_normalize(self, X):
        self.meanX = np.nanmean(X, axis=0)
        self.stdX = np.nanstd(X, axis=0)
        if not all(list(self.stdX > 1e-15)):
            logging.warning(f"Some features have std=0, {self.stdX}")
        self.stdX = np.where(self.stdX < 1e-15, 1, self.stdX)
        return self.transform_normalize(X)

    def transform_normalize(self, X):
        return (X - self.meanX) / self.stdX

    def _preprocess_data(self, p_attr):
        groups_ids = pd.DataFrame()
        p_attr = pd.Series(p_attr)

        for g in np.unique(p_attr):
            groups_ids[f"{g}"] = p_attr == g

        group_count = p_attr.value_counts()
        group_prob = group_count / len(p_attr)

        return group_prob, groups_ids

    def fit(self, X, p_attr, random_state):

        group_prob, groups_ids = self._preprocess_data(p_attr)
        # Scale and Normalize Features
        if self.normalize_input:
            X = self.fit_transform_normalize(X)
            X = normalizefea(X)

        self.fair_clustering.fit(X, group_prob, groups_ids, random_state)

        min_balance, avg_balance = get_fair_accuracy(
            group_prob, groups_ids, self.fair_clustering.l, self.K
        )
        fairness_error = get_fair_accuracy_proportional(
            group_prob, groups_ids, self.fair_clustering.l, self.K
        )

        return self

    def predict(self, X, p_attr):
        _, groups_ids = self._preprocess_data(p_attr)
        if self.normalize_input:
            X = self.transform_normalize(X)
            X = normalizefea(X)
        return self.fair_clustering.predict(X, groups_ids)



from typing import Optional

import numpy as np

from holisticai.utils.transformers.bias import BMInprocessing as BMImp

from .algorithm import MetaFairClassifierAlgorithm
from .algorithm_utils import MFLogger
from .constraints import FalseDiscovery, StatisticalRate


class MetaFairClassifier(BMImp):
    """
    The meta algorithm here takes the fairness metric as part of the input
    and returns a classifier optimized w.r.t. that fairness metric.
    The algorithm support only binary protected groups -> group_a = 1 - group_b

    References
    ----------
        Celis, L. Elisa, et al. "Classification with fairness constraints:
        A meta-algorithm with provable guarantees." Proceedings of the conference on
        fairness, accountability, and transparency. 2019.
    """

    def __init__(
        self,
        tau: Optional[float] = 0.8,
        constraint: Optional[str] = "StatisticalRate",
        seed: Optional[int] = None,
        verbose: Optional[int] = 0,
    ):
        """
        Parameters
        ----------
            tau : float
                Fairness penalty parameter (0,1). Higher parameter increate the threshold
                for a valid estimator.

            constraint : str
                The type of fairness metric to be used, currently supported:
                - StatisticalRate: Statistical rate/disparate impact
                - FalseDiscovery: False discovery rate ratio

            seed : int
                Random seed.
        """
        # Constant parameters
        steps = 10
        eps = 0.01
        mu = 0.01

        logger = MFLogger(tau, eps=eps, steps=steps, verbose=verbose)

        if constraint == "StatisticalRate":
            constraint = StatisticalRate(mu=mu)

        elif constraint == "FalseDiscovery":
            constraint = FalseDiscovery(mu=mu)

        self.algorithm = MetaFairClassifierAlgorithm(
            constraint=constraint, tau=tau, eps=eps, steps=steps, logger=logger
        )
        self.seed = seed

    def transform_estimator(self):
        return self

    def fit(
        self,
        X: np.ndarray,
        y_true: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
    ):
        """
        Fit the model

        Description
        -----------
        Learn a fair classifier.

        Parameters
        ----------

        X : numpy array
            input matrix

        y_true : numpy array
            target vector

        group_a : numpy array
            binary mask vector

        group_b : numpy array
            binary mask vector

        Returns
        -------
        the same object
        """
        params = self._load_data(y_true=y_true, group_a=group_a, group_b=group_b)
        y_true = params["y_true"]
        group_a = params["group_a"]
        group_b = params["group_b"]
        sensitive_features = np.stack([group_a, group_b], axis=1)
        self.classes_ = params["classes_"]

        self.algorithm.fit(
            X=X, y=y_true, sensitive_features=sensitive_features, random_state=self.seed
        )
        return self

    def predict(self, X: np.ndarray):
        """
        Prediction

        Description
        ----------
        Predict output for the given samples.

        Parameters
        ----------
        X : pandas.DataFrame or numpy array
            Test samples.

        Returns
        -------

        numpy.ndarray: Predicted output per sample.
        """
        return self.algorithm.predict(X)

    def predict_proba(self, X: np.ndarray):
        """
        Probability Prediction

        Description
        ----------
        Probability estimate for the given samples.

        Parameters
        ----------
        X : pandas.DataFrame or numpy array
            Test samples.

        Returns
        -------
        numpy.ndarray
            probability output per sample.
        """
        return self.algorithm.predict_proba(X)



# Base Imports
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt

# utils
from ...utils import get_colors, mat_to_binary, normalize_tensor
from ...utils._validation import _recommender_checks


def long_tail_plot(
    mat_pred, top=None, thresh=0.5, normalize=False, ax=None, size=None, title=None
):
    """
    Long Tail Plot.

    Description
    -----------
    This function plots the counts in the predictions
    for all items.

    Parameters
    ----------
    mat_pred : matrix-like
        Matrix with shape (num_users, num_items). A recommender
        score (binary or soft pred) for each user,item pair.
    top (optional) : int
        If not None, the number of items that are shown to each user.
    thresh (optional) : float
        Threshold indicating value at which
        a given item is shown to user (if top is None).
    normalize (optional) : bool
        If True, normalises the data matrix to [0,1] range.
    ax (optional) : matplotlib axes
        Pre-existing axes for the plot
    size (optional) : (int, int)
        Size of the figure
    title (optional) : str
        Title of the figure

    Returns
    -------
    matplotlib ax
    """
    # input checks and coerce
    _, _, mat_pred, _, top, thresh, normalize = _recommender_checks(
        group_a=None,
        group_b=None,
        mat_pred=mat_pred,
        mat_true=None,
        top=top,
        thresh=thresh,
        normalize=normalize,
    )

    # normalize
    if normalize:
        mat_pred = normalize_tensor(mat_pred)

    # make binary (ie shown / not shown)
    binary_mat_pred = mat_to_binary(mat_pred, top=top, thresh=thresh)

    # item counts and sort
    item_counts = list(binary_mat_pred.sum(axis=0))
    item_counts_sorted = sorted(item_counts, reverse=True)

    # setup
    sns.set()
    if ax is None:
        fig, ax = plt.subplots(figsize=size)
        if title is not None:
            fig.suptitle(title)
        else:
            fig.suptitle("Long Tail Plot")

    # chart
    ax.set_xlabel("Items (sorted by popularity)")
    ax.set_ylabel("Item Count")
    hai_color = get_colors(1, extended_colors=False)
    ax.plot(
        range(len(item_counts_sorted)),
        item_counts_sorted,
        linewidth=2,
        color=hai_color[0],
    )

    return ax


def exposure_diff_plot(
    group_a,
    group_b,
    mat_pred,
    top=None,
    thresh=0.5,
    normalize=False,
    ax=None,
    size=None,
    title=None,
):
    """
    Exposure Difference plot.

    Description
    -----------
    This function plots the difference in the exposure
    distributions between group_a and group_b.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    mat_pred : matrix-like
        Matrix with shape (num_users, num_items). A recommender
        score (binary or soft pred) for each user,item pair.
    top (optional) : int
        If not None, the number of items that are shown to each user.
    thresh (optional) : float
        Threshold in (0,1) range indicating value at which
        a given item is shown to user (if top is None).
    normalize (optional) : bool
        If True, normalises the data matrix to [0,1] range.
    ax (optional) : matplotlib axes
        Pre-existing axes for the plot
    size (optional) : (int, int)
        Size of the figure
    title (optional) : str
        Title of the figure

    Returns
    -------
    matplotlib ax
    """
    # input checks and coerce
    group_a, group_b, mat_pred, _, top, thresh, normalize = _recommender_checks(
        group_a=group_a,
        group_b=group_b,
        mat_pred=mat_pred,
        mat_true=None,
        top=top,
        thresh=thresh,
        normalize=normalize,
    )

    # normalise
    if normalize:
        mat_pred = normalize_tensor(mat_pred)

    # make binary (ie shown / not shown)
    binary_mat_pred = mat_to_binary(mat_pred, top=top, thresh=thresh)

    # Split by group
    mat_pred_a = binary_mat_pred[group_a == 1]
    mat_pred_b = binary_mat_pred[group_b == 1]

    # Get the item exposure distribution for min
    item_count_a = np.nansum(mat_pred_a, axis=0)
    item_dist_a = item_count_a / item_count_a.sum()

    # Get the item exposure distribution for maj
    item_count_b = np.nansum(mat_pred_b, axis=0)
    item_dist_b = item_count_b / item_count_b.sum()

    # take difference
    item_dist_diff = item_dist_a - item_dist_b

    # sort
    item_dist_diff_sorted = sorted(item_dist_diff, reverse=True)
    item_dist_diff_sorted = list(item_dist_diff_sorted)

    # setup
    sns.set()
    if ax is None:
        fig, ax = plt.subplots(figsize=size)
        if title is not None:
            fig.suptitle(title)
        else:
            fig.suptitle("Exposure Difference Plot")

    # chart
    ax.set_xlabel("Items (sorted by exposure difference)")
    ax.set_ylabel("$Exposure_a - Exposure_b$")
    hai_color = get_colors(1, extended_colors=False)
    ax.plot(
        range(len(item_dist_diff_sorted)),
        item_dist_diff_sorted,
        linewidth=2,
        color=hai_color[0],
    )

    return ax


def exposure_ratio_plot(
    group_a,
    group_b,
    mat_pred,
    top=None,
    thresh=0.5,
    normalize=False,
    ax=None,
    size=None,
    title=None,
):
    """
    Exposure Ratio plot.

    Description
    -----------
    This function plots the ratio in the exposure
    distributions between group_a and group_b.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    mat_pred : matrix-like
        Matrix with shape (num_users, num_items). A recommender
        score (binary or soft pred) for each user,item pair.
    top (optional) : int
        If not None, the number of items that are shown to each user.
    thresh (optional) : float
        Threshold in (0,1) range indicating value at which
        a given item is shown to user (if top is None).
    normalize (optional) : bool
        If True, normalises the data matrix to [0,1] range.
    ax (optional) : matplotlib axes
        Pre-existing axes for the plot
    size (optional) : (int, int)
        Size of the figure
    title (optional) : str
        Title of the figure

    Returns
    -------
    matplotlib ax
    """
    # input checks and coerce
    group_a, group_b, mat_pred, _, top, thresh, normalize = _recommender_checks(
        group_a=group_a,
        group_b=group_b,
        mat_pred=mat_pred,
        mat_true=None,
        top=top,
        thresh=thresh,
        normalize=normalize,
    )

    # normalise
    if normalize:
        mat_pred = normalize_tensor(mat_pred)

    # make binary (ie shown / not shown)
    binary_mat_pred = mat_to_binary(mat_pred, top=top, thresh=thresh)

    # Split by group
    mat_pred_a = binary_mat_pred[group_a == 1]
    mat_pred_b = binary_mat_pred[group_b == 1]

    # Get the item exposure distribution for min
    item_count_a = np.nansum(mat_pred_a, axis=0)
    item_dist_a = item_count_a / item_count_a.sum()

    # Get the item exposure distribution for maj
    item_count_b = np.nansum(mat_pred_b, axis=0)
    item_dist_b = item_count_b / item_count_b.sum()

    # take ratio
    item_dist_rat = item_dist_a / item_dist_b

    # sort by absolute value
    item_dist_rat_sorted = sorted(item_dist_rat, reverse=True)
    item_dist_rat_sorted = list(item_dist_rat_sorted)

    # setup
    sns.set()
    if ax is None:
        fig, ax = plt.subplots(figsize=size)
        if title is not None:
            fig.suptitle(title)
        else:
            fig.suptitle("Exposure Ratio Plot")

    # chart
    ax.set_xlabel("Items (sorted by exposure ratio)")
    ax.set_ylabel("$Exposure_a/Exposure_b$")
    hai_color = get_colors(1, extended_colors=False)
    ax.plot(
        range(len(item_dist_rat_sorted)),
        item_dist_rat_sorted,
        linewidth=2,
        color=hai_color[0],
    )

    return ax



import numpy as np
from scipy import sparse
from sklearn.metrics.pairwise import euclidean_distances as ecdist


def reduce_func(D_chunk, start):
    J = np.mean(D_chunk, axis=1)
    return J


def NormalizedCutEnergy(A, S, clustering):
    if isinstance(A, np.ndarray):
        d = np.sum(A, axis=1)

    elif isinstance(A, sparse.csc_matrix):

        d = A.sum(axis=1)

    maxclusterid = np.max(clustering)
    nassoc_e = 0
    num_cluster = 0
    for k in range(maxclusterid + 1):
        S_k = S[:, k]
        # print S_k
        if 0 == np.sum(clustering == k):
            continue  # skip empty cluster
        num_cluster = num_cluster + 1
        if isinstance(A, np.ndarray):
            nassoc_e = nassoc_e + np.dot(np.dot(np.transpose(S_k), A), S_k) / np.dot(
                np.transpose(d), S_k
            )
        elif isinstance(A, sparse.csc_matrix):
            nassoc_e = nassoc_e + np.dot(np.transpose(S_k), A.dot(S_k)) / np.dot(
                np.transpose(d), S_k
            )
            nassoc_e = nassoc_e[0, 0]
    ncut_e = num_cluster - nassoc_e

    return ncut_e


def NormalizedCutEnergy_discrete(A, clustering):
    if isinstance(A, np.ndarray):
        d = np.sum(A, axis=1)

    elif isinstance(A, sparse.csc_matrix):

        d = A.sum(axis=1)

    maxclusterid = np.max(clustering)
    nassoc_e = 0
    num_cluster = 0
    for k in range(maxclusterid + 1):
        S_k = np.array(clustering == k, dtype=np.float)
        # print S_k
        if 0 == np.sum(clustering == k):
            continue  # skip empty cluster
        num_cluster = num_cluster + 1
        if isinstance(A, np.ndarray):
            nassoc_e = nassoc_e + np.dot(np.dot(np.transpose(S_k), A), S_k) / np.dot(
                np.transpose(d), S_k
            )
        elif isinstance(A, sparse.csc_matrix):
            nassoc_e = nassoc_e + np.dot(np.transpose(S_k), A.dot(S_k)) / np.dot(
                np.transpose(d), S_k
            )
            nassoc_e = nassoc_e[0, 0]
    ncut_e = num_cluster - nassoc_e

    return ncut_e


def KernelBound_k(A, d, S_k, N):
    # S_k = S[:,k]
    volume_s_k = np.dot(np.transpose(d), S_k)
    volume_s_k = volume_s_k[0, 0]
    temp = np.dot(np.transpose(S_k), A.dot(S_k)) / (volume_s_k * volume_s_k)
    temp = temp * d
    temp2 = temp + np.reshape(-2 * A.dot(S_k) / volume_s_k, (N, 1))

    return temp2.flatten()


def km_le(X, M):

    """
    Discretize the assignments based on center

    """
    e_dist = ecdist(X, M)
    l = e_dist.argmin(axis=1)

    return l


# Fairness term calculation
def fairness_term_V_j(u_j, S, V_j):
    V_j = V_j.astype("float")
    S_term = np.maximum(np.dot(V_j, S), 1e-20)
    S_sum = np.maximum(S.sum(0), 1e-20)
    S_term = u_j * (np.log(S_sum) - np.log(S_term))
    return S_term


def km_discrete_energy(e_dist, l, k):
    tmp = np.asarray(np.where(l == k)).squeeze()
    return np.sum(e_dist[tmp, k])


def compute_fairness_energy(group_prob, groups_ids, S, bound_lambda):
    """
    compute fair clustering energy
    """
    fairness_E = [
        fairness_term_V_j(group_prob.loc[g], S, groups_ids[f"{g}"])
        for g in group_prob.index
    ]
    fairness_E = (bound_lambda * sum(fairness_E)).sum()
    return fairness_E


def _is_arraylike(x):
    """Returns whether the input is array-like."""
    return hasattr(x, "__len__") or hasattr(x, "shape") or hasattr(x, "__array__")


def _is_arraylike_not_scalar(array):
    """Return True if array is array-like and not a scalar"""
    return _is_arraylike(array) and not np.isscalar(array)


def _init_centroids(
    X, n_clusters, init, random_state=np.random.RandomState(42), init_size=None
):

    from sklearn.cluster._kmeans import _kmeans_plusplus
    from sklearn.utils.validation import check_array

    x_squared_norms = np.linalg.norm(X, axis=1)
    n_samples = X.shape[0]
    # n_clusters = n_clusters if n_centroids is None else n_centroids

    if init_size is not None and init_size < n_samples:
        init_indices = random_state.randint(0, n_samples, init_size)
        X = X[init_indices]
        x_squared_norms = x_squared_norms[init_indices]
        n_samples = X.shape[0]

    if isinstance(init, str) and init == "k-means++":
        centers, _ = _kmeans_plusplus(
            X,
            n_clusters,
            random_state=random_state,
            x_squared_norms=x_squared_norms,
        )
    elif isinstance(init, str) and init == "random":
        seeds = random_state.permutation(n_samples)[:n_clusters]
        centers = X[seeds]
    elif _is_arraylike_not_scalar(init):
        centers = init
    elif callable(init):
        centers = init(X, n_clusters, random_state=random_state)
        centers = check_array(centers, dtype=X.dtype, copy=False, order="C")
        # _validate_center_shape(X, centers)
    import scipy.sparse as sp

    if sp.issparse(centers):
        centers = centers.toarray()

    return centers



from typing import Optional

import numpy as np
import pandas as pd

from holisticai.utils.transformers.bias import BMPreprocessing as BMPre
from holisticai.utils.transformers.bias import SensitiveGroups


class Reweighing(BMPre):
    """
    Reweighing preprocessing weights the examples in each group-label combination to ensure fairness before
    classification.

    References
    ----------
        Kamiran, Faisal, and Toon Calders. "Data preprocessing techniques for classification
        without discrimination." Knowledge and information systems 33.1 (2012): 1-33.
    """

    def __init__(self):
        self.sens_groups = SensitiveGroups()

    def fit(
        self,
        y_true: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
        sample_weight: Optional[np.ndarray] = None,
    ):
        """
        Fit.

        Description
        ----------
        Access fitted sample_weight param with self.estimator_params["sample_weight"].

        Parameters
        ----------
        y_true : array-like
            Target vector
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)
        sample_weight (optional) : array-like
            Samples weights vector

        Returns
        -------
        Self
        """

        params = self._load_data(
            y_true=y_true, sample_weight=sample_weight, group_a=group_a, group_b=group_b
        )
        y_true = params["y_true"]
        sample_weight = params["sample_weight"]
        group_a = params["group_a"]
        group_b = params["group_b"]

        group_lbs = self.sens_groups.fit_transform(np.stack([group_a, group_b], axis=1))

        classes = np.unique(y_true)

        df = pd.DataFrame()

        df["LABEL"] = pd.Series(y_true)

        df["GROUP_ID"] = group_lbs

        df["COUNT"] = 1

        for g in self.sens_groups.group_names:
            for c in classes:
                df[f"{g}-{c}"] = (df["GROUP_ID"] == g) & (df["LABEL"] == c)

        df_group_values = df.groupby(["GROUP_ID", "LABEL"])["COUNT"].sum()

        df_values = df_group_values.groupby(level="LABEL").sum()

        df_groups = df_group_values.groupby(level="GROUP_ID").sum()

        df_group_values_prob = df_group_values / df_groups

        df_values_prob = df_values / df_values.sum()

        df_group_values_weights = df_values_prob / df_group_values_prob

        sample_weight = np.ones_like(y_true, dtype=np.float32)
        for g in self.sens_groups.group_names:
            for l in classes:
                mask = df[f"{g}-{l}"]
                sample_weight[mask] = df_group_values_weights.at[g, l]

        self.update_estimator_param("sample_weight", sample_weight)

        return self

    def transform(self, X: np.ndarray):
        """passthrough"""
        return X

    def fit_transform(
        self,
        X: np.ndarray,
        y_true: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
        sample_weight: Optional[np.ndarray] = None,
    ):
        """
        Fit transform.

        Description
        ----------
        Access fitted sample_weight param with self.estimator_params["sample_weight"].
        The transform returns the same object inputed.

        Parameters
        ----------
        X : matrix-like
            Input matrix
        y_true : array-like
            Target vector
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)
        sample_weight (optional) : array-like
            Samples weights vector

        Returns
        -------
            X
        """
        return self.fit(y_true, group_a, group_b, sample_weight).transform(X)



"""
The :mod:`holisticai.bias.metrics` module includes classification, regression, multiclass, recommender and clustering bias metrics
"""

# Classification
from ._classification import (
    abroca,
    accuracy_diff,
    average_odds_diff,
    classification_bias_metrics,
    cohen_d,
    disparate_impact,
    equal_opportunity_diff,
    false_negative_rate_diff,
    false_positive_rate_diff,
    four_fifths,
    statistical_parity,
    success_rate,
    true_negative_rate_diff,
    z_test_diff,
    z_test_ratio,
)

# Clustering
from ._clustering import (
    cluster_balance,
    cluster_dist_entropy,
    cluster_dist_kl,
    cluster_dist_l1,
    clustering_bias_metrics,
    min_cluster_ratio,
    silhouette_diff,
    social_fairness_ratio,
)

# Multiclass
from ._multiclass import (
    accuracy_matrix,
    confusion_matrix,
    confusion_tensor,
    frequency_matrix,
    multiclass_average_odds,
    multiclass_bias_metrics,
    multiclass_equality_of_opp,
    multiclass_statistical_parity,
    multiclass_true_rates,
    precision_matrix,
    recall_matrix,
)

# Recommender
from ._recommender import (
    aggregate_diversity,
    avg_f1_ratio,
    avg_precision_ratio,
    avg_recall_ratio,
    avg_recommendation_popularity,
    exposure_entropy,
    exposure_kl,
    exposure_l1,
    gini_index,
    mad_score,
    recommender_bias_metrics,
    recommender_mae_ratio,
    recommender_rmse_ratio,
)

# Regression
from ._regression import (
    avg_score_diff,
    avg_score_ratio,
    correlation_diff,
    disparate_impact_regression,
    mae_ratio,
    max_statistical_parity,
    no_disparate_impact_level,
    regression_bias_metrics,
    rmse_ratio,
    statistical_parity_auc,
    statistical_parity_regression,
    success_rate_regression,
    zscore_diff,
)

# All bias functions and classes
__all__ = [
    "statistical_parity",
    "disparate_impact",
    "four_fifths",
    "cohen_d",
    "accuracy_diff",
    "false_negative_rate_diff",
    "true_negative_rate_diff",
    "abroca",
    "statistical_parity_regression",
    "disparate_impact_regression",
    "success_rate_regression",
    "success_rate",
    "no_disparate_impact_level",
    "avg_score_diff",
    "avg_score_ratio",
    "zscore_diff",
    "max_statistical_parity",
    "statistical_parity_auc",
    "correlation_diff",
    "rmse_ratio",
    "mae_ratio",
    "classification_bias_metrics",
    "regression_bias_metrics",
    "aggregate_diversity",
    "gini_index",
    "avg_recommendation_popularity",
    "exposure_l1",
    "exposure_kl",
    "mad_score",
    "exposure_entropy",
    "cluster_balance",
    "min_cluster_ratio",
    "cluster_dist_l1",
    "cluster_dist_kl",
    "cluster_dist_entropy",
    "social_fairness_ratio",
    "avg_precision_ratio",
    "avg_recall_ratio",
    "avg_f1_ratio",
    "recommender_rmse_ratio",
    "recommender_mae_ratio",
    "equal_opportunity_diff",
    "false_positive_rate_diff",
    "average_odds_diff",
    "silhouette_diff",
    "recommender_bias_metrics",
    "clustering_bias_metrics",
    "confusion_matrix",
    "frequency_matrix",
    "confusion_tensor",
    "accuracy_matrix",
    "precision_matrix",
    "recall_matrix",
    "multiclass_equality_of_opp",
    "multiclass_average_odds",
    "multiclass_true_rates",
    "multiclass_statistical_parity",
    "multiclass_bias_metrics",
    "z_test_diff",
    "z_test_ratio",
]
